#include "tcaps_ci.h"

/* Set our housekeeping pointer to NULL and tell the user we exist.  */
TCAPS_Network_Task::TCAPS_Network_Task (ACE_SOCK_Dgram * udpsock)
  : n_threads_ (1)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Network_Task ctor 0x%x\n", (void *) this));
	udp_socket = udpsock;
}

/* Take care of cleanup & tell the user we're going away.  */
TCAPS_Network_Task::~TCAPS_Network_Task (void)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Network_Task dtor 0x%x\n", (void *) this));

  /* Get our shutdown notification out of the queue and release it.  */
  ACE_Message_Block *message;

  /* Like the getq() in svc() below, this will block until a message
    arrives.  By blocking, we know that the destruction will be paused
    until the last thread is done with the message block.  */
  this->getq (message);
  message->release();
  
  if(udp_socket != NULL)
  	delete udp_socket;
}

/* Open the object to do work.  Next, we activate the Task into the
  number of requested threads.  */
int TCAPS_Network_Task::open (void *unused)
{
  ACE_UNUSED_ARG (unused);

  return this->activate (THR_NEW_LWP, n_threads_);
}

/* Tell the user we're closing and invoke the baseclass' close() to
  take care of things.  */
int TCAPS_Network_Task::close (u_long flags)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Network_Task close 0x%x\n", (void *) this));
  return inherited::close (flags);
}

/* Our svc() method waits for work on the queue and then processes
  that work.  */
int TCAPS_Network_Task::svc (void)
{
	int retval = 0, num_serialized_bytes = 0;
	char buf[BUFFER_SIZE];
	ACE_Message_Block * block;
	ACE_INET_Addr remote;
	ACE_INET_Addr local;
	ACE_Time_Value wait_for_message(0, 100); //0.1ms
	struct tcaps_sp_packet packet;
	
	while(1)
	{
		//let's check if there's any receiving to be done for this socket
		//recv (void *buf, size_t n, ACE_Addr &addr, int flags, const ACE_Time_Value *timeout) const
		udp_socket->get_local_addr(local);
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Checking for packets to receive on port %d!\n", local.get_port_number()));
		retval = udp_socket->recv(buf, sizeof (buf), remote, 0, &wait_for_message);
		
		if(retval != -1)
		{
			//We received a packet!
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Num bytes (%d) received on port %d from (%s:%d)\n", retval, local.get_port_number(), remote.get_host_addr(), remote.get_port_number()));
			
			tcaps_sp_parse_packet((u8 *)buf, retval, &packet);
			
			block = handle_received_packet(&packet, &remote);
			
			if(block != NULL)
			{
				//stick the packet at the end of the inbound queue
				inbound.enqueue_tail(block);
			}
		}
		
		//let's check if there are any packets to transmit via this socket
		retval = outbound.dequeue_head(block, &wait_for_message);
		
		if(retval != -1)
		{
			//We have a packet to transmit!
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) about to read a message from the outbound_queue!\n"));
  			ACE_OS::memmove ((char *) &remote, block->rd_ptr(), sizeof (remote));
  			block->rd_ptr (sizeof (remote));  // Move the rd_ptr() beyond the data.
  			
  			ACE_OS::memmove ((char *) &num_serialized_bytes, block->rd_ptr(), sizeof (unsigned int));
  			block->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
  			
  			if(num_serialized_bytes > 0)
  			{
	  			u8 serialized_bytes[num_serialized_bytes];
	  			
	  			ACE_OS::memmove ((char *) &serialized_bytes, block->rd_ptr(), num_serialized_bytes);
	  			block->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
				
				retval = udp_socket->send(&serialized_bytes, num_serialized_bytes, remote);
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Num bytes (%d) sent to (%s:%d)\n", retval, remote.get_host_addr(), remote.get_port_number()));
  			}
			
			block->release();
		}
		
		
		//let's check if there are any messages in the task's signalling queue
		retval = this->getq(block, &wait_for_message);
		
		if(retval != -1)
		{
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) There is a message in the signalling queue!\n"));
			
			if(block->msg_type() == ACE_Message_Block::MB_HANGUP)
        	{
        		ACE_INET_Addr local;
        		udp_socket->get_local_addr(local);
        		ACE_DEBUG ((LM_DEBUG, "(%P|%t) We've been told to shutdown the network thread using port %d\n", local.get_port_number()));
          		this->putq(block);
          		break;
        	}
		}
	}
	
	return 0;
}

Packet_Block * TCAPS_Network_Task::handle_received_packet(struct tcaps_sp_packet * packet, ACE_INET_Addr * remote)
{
	Packet_Block * rx;
	unsigned int static_block_bytes = sizeof(ACE_INET_Addr) + sizeof(unsigned int);

	switch(packet->packet_type)
	{
		case TCAPS_RMP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_rmp * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_rmp * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_rmp * )packet->packet)->serialized, ((struct tcaps_sp_rmp * )packet->packet)->serialized_bytes);
			tcaps_sp_rmp_destroy((struct tcaps_sp_rmp * )packet->packet);
			break;
		case TCAPS_AP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_ap * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_ap * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_ap * )packet->packet)->serialized, ((struct tcaps_sp_ap * )packet->packet)->serialized_bytes);
			tcaps_sp_ap_destroy((struct tcaps_sp_ap * )packet->packet);
			break;
		case TCAPS_TP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_tp * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_tp * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_tp * )packet->packet)->serialized, ((struct tcaps_sp_tp * )packet->packet)->serialized_bytes);
			tcaps_sp_tp_destroy((struct tcaps_sp_tp * )packet->packet);
			break;
		case TCAPS_PP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_pp * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_pp * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_pp * )packet->packet)->serialized, ((struct tcaps_sp_pp * )packet->packet)->serialized_bytes);
			tcaps_sp_pp_destroy((struct tcaps_sp_pp * )packet->packet);
			break;
		case TCAPS_FFP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_ffp * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_ffp * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_ffp * )packet->packet)->serialized, ((struct tcaps_sp_ffp * )packet->packet)->serialized_bytes);
			tcaps_sp_ffp_destroy((struct tcaps_sp_ffp * )packet->packet);
			break;
		case TCAPS_CRMP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_crmp * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_crmp * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_crmp * )packet->packet)->serialized, ((struct tcaps_sp_crmp * )packet->packet)->serialized_bytes);
			tcaps_sp_crmp_destroy((struct tcaps_sp_crmp * )packet->packet);
			break;
		case TCAPS_CMP_PACKET_TYPE:
			rx = new Packet_Block(static_block_bytes + ((struct tcaps_sp_cmp * )packet->packet)->serialized_bytes);
  			rx->copy ((const char *) remote, sizeof(ACE_INET_Addr));
  			rx->copy ((const char *) &(((struct tcaps_sp_cmp * )packet->packet)->serialized_bytes), sizeof(unsigned int));
  			rx->copy ((const char *) ((struct tcaps_sp_cmp * )packet->packet)->serialized, ((struct tcaps_sp_cmp * )packet->packet)->serialized_bytes);
			tcaps_sp_cmp_destroy((struct tcaps_sp_cmp * )packet->packet);
			break;
		default:
			rx = NULL;
			break;
	}
	
	return rx;
}

void TCAPS_Network_Task::send_packet(Packet_Block * tosend)
{
	outbound.enqueue_tail(tosend);
}

Packet_Block * TCAPS_Network_Task::receive_packet()
{
	ACE_Time_Value wait_for_message(0, 100); //0.1ms
	ACE_Message_Block * rx;
	int retval = 0;

	retval = inbound.dequeue_head(rx, &wait_for_message);
	
	if(retval == -1)
	{
		return NULL;
	}
	
	return (Packet_Block *)rx;
}
