#include "tcaps_fs.h"

static void pad(char * string, unsigned int len, unsigned int n);

TCAPS_Flow::TCAPS_Flow (TCAPS_Flow_ID * id)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow ctor 0x%x\n", (void *) this));
	num_packets_in_window = 0;
	flow_start_time_secs = (unsigned int)ACE_OS::time();
	last_rx_packet = NULL;
	last_rx_client2other_packet = NULL;
	last_rx_other2client_packet = NULL;
	flow_state_changed = false;
	
	flow_id = id;
	
	num_payload_bytes_bidir_since_beginning = 0;
	num_payload_bytes_client2other_since_beginning = 0;
	num_payload_bytes_other2client_since_beginning = 0;
	
	num_packets_bidir_since_beginning = 0;
	num_packets_client2other_since_beginning = 0;
	num_packets_other2client_since_beginning = 0;
	
	average_interarrival_time_bidir_ms = 0;
	average_interarrival_time_client2other_ms = 0;
	average_interarrival_time_other2client_ms = 0;
	
	min_interarrival_time_client2other_ms = 0;
	min_interarrival_time_other2client_ms = 0;
	max_interarrival_time_client2other_ms = 0;
	max_interarrival_time_other2client_ms = 0;
	
	min_client2other_payload_bytes = 0;
	min_other2client_payload_bytes = 0;
	max_client2other_payload_bytes = 0;
	max_other2client_payload_bytes = 0;
}

TCAPS_Flow::TCAPS_Flow (const TCAPS_Flow & copied_flow)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow copy ctor 0x%x\n", (void *) this));
	flow_packets = copied_flow.get_packet_list();
	flow_id = copied_flow.get_flow_id();
	num_packets_in_window = copied_flow.get_num_packets_in_flow();
}

TCAPS_Flow::~TCAPS_Flow (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow dtor 0x%x\n", (void *) this)); 
	
	if(flow_id != NULL)
		delete flow_id;
		
	std::list<Pcap_Packet*>::iterator i;
	
	for(i = flow_packets.begin(); i != flow_packets.end(); i++)
	{
		delete *i;
	}
}

const Pcap_Packet * TCAPS_Flow::get_last_rx_packet() const
{
	return last_rx_packet;
}
	
	
bool operator<(const TCAPS_Flow &f1, const TCAPS_Flow & f2)
{
	return (f1.get_flow_id() < f2.get_flow_id());
}

bool operator>(const TCAPS_Flow &f1, const TCAPS_Flow & f2)
{
	return (f1.get_flow_id() > f2.get_flow_id());
}

bool operator==(const TCAPS_Flow &f1, const TCAPS_Flow & f2)
{
	return (f1.get_flow_id() == f2.get_flow_id());
}


TCAPS_Flow & TCAPS_Flow::operator=(const TCAPS_Flow & copied_flow)
{
	if(this == &copied_flow)
		return *this;
		
	flow_packets = copied_flow.get_packet_list();
	flow_id = copied_flow.get_flow_id();
	num_packets_in_window = copied_flow.get_num_packets_in_flow();
	
	return *this;
}

bool TCAPS_Flow::state_changed() const
{
	return flow_state_changed;
}

void TCAPS_Flow::reset_state_changed()
{
	flow_mutex.acquire();
	flow_state_changed = false;
	flow_mutex.release();
}

unsigned int TCAPS_Flow::get_num_packets_in_flow() const
{
	return num_packets_in_window;
}

std::list<Pcap_Packet*> TCAPS_Flow::get_packet_list() const
{
	return flow_packets;
}


TCAPS_Flow_ID * TCAPS_Flow::get_flow_id() const
{
	return flow_id;
}

bool TCAPS_Flow::is_client_to_other_packet(Pcap_Packet * packet, const struct ip_address * clientip)
{
	char ipbuf1[60], ipbuf2[60], ipbuf3[60];
	//we assume that the layer 3 protocol is IP and that this has been enforced elsewhere
	
	if(packet->get_layer2_frame()->get_layer3_packet()->get_layer3_protocol() == LAYER3_PROTOCOL_IPV4)
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet we are about to add is IPv4\n"));
		
		tcaps_ip_ntoa(clientip, ipbuf1, TRUE);
		tcaps_ip_ntoa(((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip(), ipbuf2, TRUE);
		tcaps_ip_ntoa(((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_dst_ip(), ipbuf3, TRUE);
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) clientip: %s\tpacket src ip: %s\tpacket dst ip: %s\n", ipbuf1, ipbuf2, ipbuf3));
		
		if(memcmp(clientip->octets, ((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV4_OCTETS) == 0)
		{
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
			
			return true;
		}
		else
		{
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
			
			return false;
		}
	}
	else //LAYER3_PROTOCOL_IPV6
	{
		if(memcmp(clientip->octets, ((IPv6_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV6_OCTETS) == 0)
		{
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
			
			return true;
		}
		else
		{
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
			
			return false;
		}
	}
}

void TCAPS_Flow::add_packet(Pcap_Packet * packet, const struct ip_address * clientip)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to add a packet to a flow\n")); 
	
	bool is_client2other_packet = is_client_to_other_packet(packet, clientip);

	if(num_packets_bidir_since_beginning == 0)
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Updating last_rx_packet pointer for the first time\n")); 
		last_rx_packet = new Pcap_Packet(*packet);
	}
	else
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Updating last_rx_packet pointer\n")); 
		delete last_rx_packet;
		last_rx_packet = new Pcap_Packet(*flow_packets.back());
	}
	
	//let's signal that this flow is locked just while we add packets
	//used to juggle concurrency between flow_manager thread and flow_feature thread
	#ifdef _MUTEX_DEBUG_
	#endif
	
	flow_packets_mutex.acquire();
	
	#ifdef _MUTEX_DEBUG_
	#endif
	
	while(num_packets_in_window >= MAX_PACKETS_PER_FLOW)
	{
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to insert a packet in a FULL flow\n"));
		
		//we've reached or surpassed the limit for the number of packets to keep for a flow
		//lets delete the oldest packet from the flow and append the new packet to the end
		
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) list.size() before delete: %d\n", flow_packets.size()));
	
		//Call delete the PCAP_packet * stored in the first list element
		delete(*flow_packets.begin());
		
		//Erase the first list element itself
		flow_packets.erase(flow_packets.begin());
		
		num_packets_in_window--;
		
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) list.size() after delete: %d\n", flow_packets.size()));
		
		
	}
//	else
//	{
//		
//		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to insert a packet in a NON FULL flow...num_packets: %d,    max_packets: %d\n", num_packets, MAX_PACKETS_PER_FLOW));
//	}


	//we're free to add the packet to the list
	flow_packets.push_back(packet);
	num_packets_in_window++;

	
	
	//adjust "since flow beginning" parameters
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to enter calculate_since_beginning_flow_features()\n"));
	calculate_since_beginning_flow_features(packet, is_client2other_packet);
	
	flow_packets_mutex.release();
	
	#ifdef _MUTEX_DEBUG_
	#endif
	
	flow_mutex.acquire();
	flow_state_changed = true;
	flow_mutex.release();
	
	if(is_client2other_packet)
	{
		if(last_rx_client2other_packet != NULL)
				delete last_rx_client2other_packet;
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Updating last_rx_client2other_packet pointer\n"));
		last_rx_client2other_packet = new Pcap_Packet(*packet);
	}
	else
	{
		if(last_rx_other2client_packet != NULL)
				delete last_rx_other2client_packet;
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Updating last_rx_other2client_packet pointer\n"));
		last_rx_other2client_packet = new Pcap_Packet(*packet);
	}
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished adding packet to flow\n"));
}



void TCAPS_Flow::calculate_since_beginning_flow_features(Pcap_Packet * packet, bool is_client2other_packet)
{
	//This function gets called every time a new packet is added to the flow
	//Packets arrive in the order in which they were sniffed off the wire
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) timeval from current packet looks like this: secs: %d\tusecs: %d\n", packet->get_pcap_packet_header()->ts.tv_sec, packet->get_pcap_packet_header()->ts.tv_usec));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) timeval from last_rx_packet looks like this: secs: %d\tusecs: %d\n", last_rx_packet->get_pcap_packet_header()->ts.tv_sec, last_rx_packet->get_pcap_packet_header()->ts.tv_usec));
	
	ACE_Time_Value current = ACE_Time_Value(packet->get_pcap_packet_header()->ts);
	ACE_Time_Value previous = ACE_Time_Value(last_rx_packet->get_pcap_packet_header()->ts);
	
	current -= previous;
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) ACE_Time_Value current now looks like this: secs: %d\tusecs: %d\n", current.sec(), current.usec()));

	num_packets_bidir_since_beginning++;
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) num_packets_bidir_since_beginning: %d\n", num_packets_bidir_since_beginning));
	
	//calculate bidir avg interarrival time
	
	float recent_interarrival_time_bidir_ms = ((float)(current.sec()*1000))+(((float)current.usec())/1000);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) recent_interarrival_time_bidir_ms: %0.2f\n", recent_interarrival_time_bidir_ms));
		
	if(num_packets_bidir_since_beginning > 2)
	{
		average_interarrival_time_bidir_ms = ((average_interarrival_time_bidir_ms * (num_packets_bidir_since_beginning-1)) + recent_interarrival_time_bidir_ms)/num_packets_bidir_since_beginning;
	}
	else
		average_interarrival_time_bidir_ms = recent_interarrival_time_bidir_ms;
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) average_interarrival_time_bidir_ms: %0.2f\n", average_interarrival_time_bidir_ms));
	
	num_payload_bytes_bidir_since_beginning += packet->get_payload_len();
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) num_payload_bytes_bidir_since_beginning: %d\n", num_payload_bytes_bidir_since_beginning));
	
	unsigned int total_active_time_secs = ((unsigned int)ACE_OS::time()) - flow_start_time_secs;
	
	unsigned int current_payload_len = packet->get_payload_len();
	
	if(is_client2other_packet)
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) is_client2other_packet\n"));
		
		num_packets_client2other_since_beginning++;
		num_payload_bytes_client2other_since_beginning += packet->get_payload_len();
		
		//calculate client2other avg interarrival time
		current = ACE_Time_Value(packet->get_pcap_packet_header()->ts);
		
		//if this is the first client2other packet we are processing, last_rx_other2client_packet == NULL, so we just take the current packet to be the last
		if(last_rx_client2other_packet != NULL)
			previous = ACE_Time_Value(last_rx_client2other_packet->get_pcap_packet_header()->ts);
		else
			previous = ACE_Time_Value(packet->get_pcap_packet_header()->ts);
			
		current -= previous;
		float recent_interarrival_time_client2other_ms = (current.sec()*1000)+((float)current.usec()/1000);
		
		if(num_packets_client2other_since_beginning > 2)
		{
			average_interarrival_time_client2other_ms = ((average_interarrival_time_client2other_ms * (num_packets_client2other_since_beginning-1)) + recent_interarrival_time_client2other_ms)/((float)num_packets_client2other_since_beginning);
		}
		else
			average_interarrival_time_client2other_ms = recent_interarrival_time_client2other_ms;
		
		if(recent_interarrival_time_client2other_ms < min_interarrival_time_client2other_ms || num_packets_client2other_since_beginning == 2)
			min_interarrival_time_client2other_ms = recent_interarrival_time_client2other_ms;
			
		if(recent_interarrival_time_client2other_ms > max_interarrival_time_client2other_ms  || num_packets_client2other_since_beginning == 2)
			max_interarrival_time_client2other_ms = recent_interarrival_time_client2other_ms;
	
		
		if(current_payload_len < min_client2other_payload_bytes || num_packets_client2other_since_beginning == 1)
			min_client2other_payload_bytes = current_payload_len;
		
		if(current_payload_len > max_client2other_payload_bytes  || num_packets_client2other_since_beginning == 1)
			max_client2other_payload_bytes = current_payload_len;
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) BEFORE INCREMENT - client2other_payload_bytes_histo[%d]: %d\n", current_payload_len, client2other_payload_bytes_histo[current_payload_len]));
		client2other_payload_bytes_histo[current_payload_len] += 1;
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) AFTER INCREMENT - client2other_payload_bytes_histo[%d]: %d\n", current_payload_len, client2other_payload_bytes_histo[current_payload_len]));
		
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) BEFORE INCREMENT - client2other_interarrival_time_ms_histo[%d]: %d\n", (unsigned int)floorf(recent_interarrival_time_client2other_ms), client2other_interarrival_time_ms_histo[(unsigned int)floorf(recent_interarrival_time_client2other_ms)]));
		client2other_interarrival_time_ms_histo[(unsigned int)floorf(recent_interarrival_time_client2other_ms)] += 1;
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) AFTER INCREMENT - client2other_interarrival_time_ms_histo[%d]: %d\n", (unsigned int)floorf(recent_interarrival_time_client2other_ms), client2other_interarrival_time_ms_histo[(unsigned int)floorf(recent_interarrival_time_client2other_ms)]));
	}
	else
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) is_other2client_packet\n"));
		
		num_packets_other2client_since_beginning++;
		num_payload_bytes_other2client_since_beginning += packet->get_payload_len();
		
		//calculate other2client avg interarrival time
		current = ACE_Time_Value(packet->get_pcap_packet_header()->ts);
		
		//if this is the first other2client packet we are processing, last_rx_other2client_packet == NULL, so we just take the current packet to be the last
		if(last_rx_other2client_packet != NULL)
			previous = ACE_Time_Value(last_rx_other2client_packet->get_pcap_packet_header()->ts);
		else
			previous = ACE_Time_Value(packet->get_pcap_packet_header()->ts);
			
		current -= previous;
		float recent_interarrival_time_other2client_ms = (current.sec()*1000)+((float)current.usec())/1000;
		
		if(num_packets_other2client_since_beginning > 2)
		{
			average_interarrival_time_other2client_ms = ((average_interarrival_time_other2client_ms * (num_packets_other2client_since_beginning-1)) + recent_interarrival_time_other2client_ms)/((float)num_packets_other2client_since_beginning);
		}
		else
			average_interarrival_time_other2client_ms = recent_interarrival_time_other2client_ms;
		
		if(recent_interarrival_time_other2client_ms < min_interarrival_time_other2client_ms || num_packets_other2client_since_beginning == 2)
			min_interarrival_time_other2client_ms = recent_interarrival_time_other2client_ms;
			
		if(recent_interarrival_time_other2client_ms > max_interarrival_time_other2client_ms || num_packets_other2client_since_beginning == 2)
			max_interarrival_time_other2client_ms = recent_interarrival_time_other2client_ms;
		
		if(current_payload_len < min_other2client_payload_bytes || num_packets_other2client_since_beginning == 1)	
			min_other2client_payload_bytes = current_payload_len;
			
		if(current_payload_len > max_other2client_payload_bytes || num_packets_other2client_since_beginning == 1)
			max_other2client_payload_bytes = current_payload_len;
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) BEFORE INCREMENT - other2client_payload_bytes_histo[%d]: %d\n", current_payload_len, other2client_payload_bytes_histo[current_payload_len]));
		other2client_payload_bytes_histo[current_payload_len] += 1;
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) AFTER INCREMENT - other2client_payload_bytes_histo[%d]: %d\n", current_payload_len, other2client_payload_bytes_histo[current_payload_len]));
		
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) BEFORE INCREMENT - other2client_interarrival_time_ms_histo[%d]: %d\n", (unsigned int)floorf(recent_interarrival_time_other2client_ms), other2client_interarrival_time_ms_histo[(unsigned int)floorf(recent_interarrival_time_other2client_ms)]));
		other2client_interarrival_time_ms_histo[(unsigned int)floorf(recent_interarrival_time_other2client_ms)] += 1;
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) AFTER INCREMENT - other2client_interarrival_time_ms_histo[%d]: %d\n", (unsigned int)floorf(recent_interarrival_time_other2client_ms), other2client_interarrival_time_ms_histo[(unsigned int)floorf(recent_interarrival_time_other2client_ms)]));
	}
	
	
	#ifdef _MUTEX_DEBUG_
	#endif
	
	flow_features_mutex.acquire();
	
	#ifdef _MUTEX_DEBUG_
	#endif
	
	//"since flow beginning" features
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS] = num_packets_bidir_since_beginning;
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PACKETS] = num_packets_client2other_since_beginning;
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PACKETS] = num_packets_other2client_since_beginning;
	
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_BIDIR_MS] = average_interarrival_time_bidir_ms;
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS] = average_interarrival_time_client2other_ms;
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS] = average_interarrival_time_other2client_ms;
	
	flow_features[TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS] = min_interarrival_time_client2other_ms;
	flow_features[TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS] = min_interarrival_time_other2client_ms;
	
	flow_features[TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS] = max_interarrival_time_client2other_ms;
	flow_features[TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS] = max_interarrival_time_other2client_ms;
	
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_ACTIVE_TIME_SECS] = total_active_time_secs;
	
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PACKETS_PER_SECOND] = divideIfNotZero((float)num_packets_bidir_since_beginning, total_active_time_secs);
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PACKETS_PER_SECOND] = divideIfNotZero((float)num_packets_client2other_since_beginning, total_active_time_secs);
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PACKETS_PER_SECOND] = divideIfNotZero((float)num_packets_other2client_since_beginning, total_active_time_secs);
	
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PAYLOAD_BYTES] = num_payload_bytes_bidir_since_beginning;
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PAYLOAD_BYTES] = num_payload_bytes_client2other_since_beginning;
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PAYLOAD_BYTES] = num_payload_bytes_other2client_since_beginning;
	
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_SECOND] = divideIfNotZero((float)num_payload_bytes_bidir_since_beginning, total_active_time_secs);
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_SECOND] = divideIfNotZero((float)num_payload_bytes_client2other_since_beginning, total_active_time_secs);
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_SECOND] = divideIfNotZero((float)num_payload_bytes_other2client_since_beginning, total_active_time_secs);
	
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_PACKET] = divideIfNotZero((float)num_payload_bytes_bidir_since_beginning, num_packets_bidir_since_beginning);
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = divideIfNotZero((float)num_payload_bytes_client2other_since_beginning, num_packets_client2other_since_beginning);
	flow_features[TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = divideIfNotZero((float)num_payload_bytes_other2client_since_beginning, num_packets_other2client_since_beginning);
	
	flow_features[TCAPS_FLOW_FEATURE_MINIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = min_client2other_payload_bytes;
	flow_features[TCAPS_FLOW_FEATURE_MINIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = min_other2client_payload_bytes;
	
	flow_features[TCAPS_FLOW_FEATURE_MAXIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = max_client2other_payload_bytes;
	flow_features[TCAPS_FLOW_FEATURE_MAXIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = max_other2client_payload_bytes;
	
	flow_features_mutex.release();
	
	#ifdef _MUTEX_DEBUG_
	#endif
	
	//char buf[2048];
	//to_string(buf, sizeof(buf));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Flow features look like this after calculation for a packet:\n%s\n", buf));
	
	
}

void TCAPS_Flow::calculate_histogram_flow_features()
{
	std::map <u8,float> client2other_payload_bytes_percentiles = calculate_histo_percentiles(client2other_payload_bytes_histo, num_packets_client2other_since_beginning);
	std::map <u8,float> other2client_payload_bytes_percentiles = calculate_histo_percentiles(other2client_payload_bytes_histo, num_packets_other2client_since_beginning);
	
	std::map <u8,float> client2other_interarrival_time_ms_percentiles = calculate_histo_percentiles(client2other_interarrival_time_ms_histo, num_packets_client2other_since_beginning);
	std::map <u8,float> other2client_interarrival_time_ms_percentiles = calculate_histo_percentiles(other2client_interarrival_time_ms_histo, num_packets_other2client_since_beginning);
	
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 5th percentile of client2other payload bytes: %0.2f\n", percentiles[5]));
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 15th percentile of client2other payload bytes: %0.2f\n", percentiles[15]));
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 25th percentile of client2other payload bytes: %0.2f\n", percentiles[25]));
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 50th percentile of client2other payload bytes: %0.2f\n", percentiles[50]));
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 75th percentile of client2other payload bytes: %0.2f\n", percentiles[75]));
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 80th percentile of client2other payload bytes: %0.2f\n", percentiles[80]));
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() 90th percentile of client2other payload bytes: %0.2f\n", percentiles[90]));
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() Acquiring flow_features_mutex\n"));
	#endif
	
	flow_features_mutex.acquire();
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() flow_features_mutex acquired\n"));
	#endif
	
	flow_features[TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = client2other_payload_bytes_percentiles[25];
	flow_features[TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = client2other_payload_bytes_percentiles[50];
	flow_features[TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = client2other_payload_bytes_percentiles[75];
	flow_features[TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = client2other_payload_bytes_percentiles[90];
	flow_features[TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET] = client2other_payload_bytes_percentiles[95];
	
	flow_features[TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = other2client_payload_bytes_percentiles[25];
	flow_features[TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = other2client_payload_bytes_percentiles[50];
	flow_features[TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = other2client_payload_bytes_percentiles[75];
	flow_features[TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = other2client_payload_bytes_percentiles[90];
	flow_features[TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET] = other2client_payload_bytes_percentiles[95];
	
	flow_features[TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET] = client2other_interarrival_time_ms_percentiles[25];
	flow_features[TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET] = client2other_interarrival_time_ms_percentiles[50];
	flow_features[TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET] = client2other_interarrival_time_ms_percentiles[75];
	flow_features[TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET] = client2other_interarrival_time_ms_percentiles[90];
	flow_features[TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET] = client2other_interarrival_time_ms_percentiles[95];
	
	flow_features[TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET] = other2client_interarrival_time_ms_percentiles[25];
	flow_features[TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET] = other2client_interarrival_time_ms_percentiles[50];
	flow_features[TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET] = other2client_interarrival_time_ms_percentiles[75];
	flow_features[TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET] = other2client_interarrival_time_ms_percentiles[90];
	flow_features[TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET] = other2client_interarrival_time_ms_percentiles[95];
	
	flow_features_mutex.release();
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histogram_flow_features() released flow_features_mutex\n"));
	#endif
}

std::map <u8,float> TCAPS_Flow::calculate_histo_percentiles(std::map <unsigned int,unsigned int> & the_histo_map, unsigned int total_samples)
{
	std::map <unsigned int,float> histo_distribution;
	std::map <unsigned int,float> histo_cumulative_distribution;
	std::map <u8,float> percentiles;
	
	std::map <unsigned int,unsigned int>::iterator histo_iter;
	std::map <unsigned int,float>::iterator dist_iter_cur, dist_iter_prev;
	
	unsigned int bucket_count = 0;
	
	for(histo_iter = the_histo_map.begin(); histo_iter != the_histo_map.end(); histo_iter++, bucket_count++)
	{
		//calculate % distribution for each of the histogram buckets
		histo_distribution[histo_iter->first] = divideIfNotZero((float)the_histo_map[histo_iter->first], total_samples);
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) calculate_histo_percentiles(): payload_bytes_bucket: %d\tdistribution_value: %f\n", histo_iter->first, histo_distribution[histo_iter->first]));
	}
	
	
	bool percentile_5th_found = false;
	bool percentile_10th_found = false;
	bool percentile_15th_found = false;
	bool percentile_20th_found = false;
	bool percentile_25th_found = false;
	bool percentile_30th_found = false;
	bool percentile_35th_found = false;
	bool percentile_40th_found = false;
	bool percentile_45th_found = false;
	bool percentile_50th_found = false;
	bool percentile_55th_found = false;
	bool percentile_60th_found = false;
	bool percentile_65th_found = false;
	bool percentile_70th_found = false;
	bool percentile_75th_found = false;
	bool percentile_80th_found = false;
	bool percentile_85th_found = false;
	bool percentile_90th_found = false;
	bool percentile_95th_found = false;
	bool percentile_100th_found = false;
	
	bool first_time = true;
	
	
	float cumulativetmp = 0;
	//dist_iter_cur = histo_distribution.begin();
	//dist_iter_prev = histo_distribution.begin();
	//histo_cumulative_distribution[dist_iter_cur->first] = histo_distribution[dist_iter_cur->first];
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) calculate_histo_percentiles(): payload_bytes_bucket: %d\tcumulative_distribution_value: %f\n", dist_iter_cur->first, histo_cumulative_distribution[dist_iter_cur->first]));
	//dist_iter_cur++;
	

	
	if(bucket_count > 1)
	{
		for(dist_iter_cur = dist_iter_prev = histo_distribution.begin(); dist_iter_cur != histo_distribution.end(); dist_iter_cur++)
		{
			if(first_time)
			{
				//first time through, we don't have a previous value in the histogram list to work with, so we assign the first bucket % to our cumulative tally
				cumulativetmp = histo_cumulative_distribution[dist_iter_cur->first] = histo_distribution[dist_iter_cur->first];
				//ACE_DEBUG ((LM_DEBUG, "(%P|%t) calculate_histo_percentiles(): payload_bytes_bucket: %d\tcumulative_distribution_value: %f\n", dist_iter_cur->first, histo_cumulative_distribution[dist_iter_cur->first]));
				first_time = false;
			}
			else
			{
				//calculate % cumulative distribution for each of the histogram buckets
				cumulativetmp = histo_cumulative_distribution[dist_iter_prev->first] + histo_distribution[dist_iter_cur->first];
				histo_cumulative_distribution[dist_iter_cur->first] = cumulativetmp;
				dist_iter_prev = dist_iter_cur;
			}
			
			if(cumulativetmp >= 0.95 && !percentile_95th_found)
			{
				percentiles[95] = dist_iter_cur->first;
				percentile_95th_found = true;
			}
			if(cumulativetmp >= 0.90 && !percentile_90th_found)
			{
				percentiles[90] = dist_iter_cur->first;
				percentile_90th_found = true;
			}
			if(cumulativetmp >= 0.85 && !percentile_85th_found)
			{
				percentiles[85] = dist_iter_cur->first;
				percentile_85th_found = true;
			}
			if(cumulativetmp >= 0.80 && !percentile_80th_found)
			{
				percentiles[80] = dist_iter_cur->first;
				percentile_80th_found = true;
			}
			if(cumulativetmp >= 0.75 && !percentile_75th_found)
			{
				percentiles[75] = dist_iter_cur->first;
				percentile_75th_found = true;
			}
			if(cumulativetmp >= 0.70 && !percentile_70th_found)
			{
				percentiles[70] = dist_iter_cur->first;
				percentile_70th_found = true;
			}
			if(cumulativetmp >= 0.65 && !percentile_65th_found)
			{
				percentiles[65] = dist_iter_cur->first;
				percentile_65th_found = true;
			}
			if(cumulativetmp >= 0.60 && !percentile_60th_found)
			{
				percentiles[60] = dist_iter_cur->first;
				percentile_60th_found = true;
			}
			if(cumulativetmp >= 0.55 && !percentile_55th_found)
			{
				percentiles[55] = dist_iter_cur->first;
				percentile_55th_found = true;
			}
			if(cumulativetmp >= 0.50 && !percentile_50th_found)
			{
				percentiles[50] = dist_iter_cur->first;
				percentile_50th_found = true;
			}
			if(cumulativetmp >= 0.45 && !percentile_45th_found)
			{
				percentiles[45] = dist_iter_cur->first;
				percentile_45th_found = true;
			}
			if(cumulativetmp >= 0.40 && !percentile_40th_found)
			{
				percentiles[40] = dist_iter_cur->first;
				percentile_40th_found = true;
			}
			if(cumulativetmp >= 0.35 && !percentile_35th_found)
			{
				percentiles[35] = dist_iter_cur->first;
				percentile_35th_found = true;
			}
			if(cumulativetmp >= 0.30 && !percentile_30th_found)
			{
				percentiles[30] = dist_iter_cur->first;
				percentile_30th_found = true;
			}
			if(cumulativetmp >= 0.25 && !percentile_25th_found)
			{
				percentiles[25] = dist_iter_cur->first;
				percentile_25th_found = true;
			}
			if(cumulativetmp >= 0.20 && !percentile_20th_found)
			{
				percentiles[20] = dist_iter_cur->first;
				percentile_20th_found = true;
			}
			if(cumulativetmp >= 0.15 && !percentile_15th_found)
			{
				percentiles[15] = dist_iter_cur->first;
				percentile_15th_found = true;
			}
			if(cumulativetmp >= 0.10 && !percentile_10th_found)
			{
				percentiles[10] = dist_iter_cur->first;
				percentile_10th_found = true;
			}
			if(cumulativetmp >= 0.05 && !percentile_5th_found)
			{
				percentiles[5] = dist_iter_cur->first;
				percentile_5th_found = true;
			}
		}
	}
	else
	{
		dist_iter_cur = histo_distribution.begin();
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histo_percentiles() only 1 histogram bucket. Assigning all percentiles to same bucket value of: %0.2f\n", (float)dist_iter_cur->first));
		percentiles[5] = percentiles[10] = percentiles[15] = percentiles[20] = percentiles[25] = (float)dist_iter_cur->first;
		percentiles[30] = percentiles[35] = percentiles[40] = percentiles[45] = percentiles[50] = (float)dist_iter_cur->first;
		percentiles[55] = percentiles[60] =	percentiles[65] = percentiles[70] = percentiles[75] = (float)dist_iter_cur->first;
		percentiles[80] = percentiles[85] = percentiles[90] = percentiles[95] = percentiles[100] = (float)dist_iter_cur->first;
	}
	
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_histo_percentiles() 25th percentile of client2other payload bytes: %0.2f\n", percentiles[25]));
	
	return percentiles;
}

void TCAPS_Flow::calculate_window_flow_features()
{

	std::list<Pcap_Packet*>::iterator i;
	unsigned int list_size=0, payload_sizes=0;
	
	//let's lock the mutex whilst we read from the flow_packets list((*dist_iter_cur).first)
	flow_packets_mutex.acquire();
	
	list_size = num_packets_in_window;
	
	for(i = flow_packets.begin(); i != flow_packets.end(); i++)
	{
		//let's calc all the stats we need to from the list
		payload_sizes += (*i)->get_payload_len();
	}
	
	//finished reading, unlock
	flow_packets_mutex.release();
	
	
	//calculate averages;
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_window_flow_features() Acquiring flow_features_mutex\n"));
	#endif
	
	flow_features_mutex.acquire();
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_window_flow_features() flow_features_mutex acquired\n"));
	#endif
	
	//sliding window parameters
	flow_features[TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS_WINDOW] = list_size;
	//flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_PAYLOAD_SIZE_BYTES_WINDOW, payload_sizes/(float)list_size));
	
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) ip_packet_sizes: %d,  num_packets: %d\n", ip_packet_sizes, list_size));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) ip_packet_sizes/(float)num_packets: %0.2f\n", ip_packet_sizes/(float)list_size));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_FLOW_FEATURE_AVERAGE_IP_PACKET_SIZE_BYTES: %0.2f\n", flow_features[TCAPS_FLOW_FEATURE_AVERAGE_IP_PACKET_SIZE_BYTES]));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_FLOW_FEATURE_AVERAGE_PAYLOAD_SIZE_BYTES: %0.2f\n", flow_features[TCAPS_FLOW_FEATURE_AVERAGE_PAYLOAD_SIZE_BYTES]));
	
	flow_features_mutex.release();
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::calculate_window_flow_features() flow_features_mutex released\n"));
	#endif
	
}

u8 * TCAPS_Flow::serialize_flow_features(unsigned int * num_serialized_features)
{
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) flow_features.size(): %d  sizeof(tcaps_flow_feature_type): %d  sizeof(tcaps_flow_feature): %d  malloc size: %d\n", flow_features.size(), sizeof(tcaps_flow_feature_type), sizeof(tcaps_flow_feature), flow_features.size()*(sizeof(tcaps_flow_feature_type)+sizeof(tcaps_flow_feature))));

	std::map<tcaps_flow_feature_type,tcaps_flow_feature>::iterator i;
	
	unsigned int count = 0;
	unsigned int offset = sizeof(tcaps_flow_feature_type) + sizeof(tcaps_flow_feature);
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::serialize_flow_features() Acquiring flow_features_mutex\n"));
	#endif
	
	flow_features_mutex.acquire();
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::serialize_flow_features() flow_features_mutex acquired\n"));
	#endif
	
	(*num_serialized_features) = flow_features.size();
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) flow_features.size(): %d (*num_serialized_features): %d\n", flow_features.size(),(*num_serialized_features)));
	unsigned int num_serialized_bytes = (*num_serialized_features)*(sizeof(tcaps_flow_feature_type)+sizeof(tcaps_flow_feature));
	u8 * serialized_features = (u8 *)malloc(num_serialized_bytes);
	
	for(i = flow_features.begin(); i != flow_features.end(); i++)
	{
		(*((tcaps_flow_feature_type*)(serialized_features+(count*offset)))) = (*i).first;
		(*((tcaps_flow_feature*)(serialized_features+sizeof(tcaps_flow_feature_type)+(count*offset)))) = (*i).second;
		count++;
	}
	
	flow_features_mutex.release();
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::serialize_flow_features() flow_features_mutex released\n"));
	#endif
	
	//char buf[1024];
	//printHexBytes(serialized_features, num_serialized_bytes, buf, sizeof(buf), FALSE);
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Contents of serialized_features in hex: %s\n", buf));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Recovered flow feature from serialized_features: %0.2f\n", *((float*)(serialized_features+2))));
	
	
	return serialized_features;
}


void TCAPS_Flow::to_string(char * buf, unsigned int bufsize)
{
	flow_id->get_ascii_flow_id(buf);

	std::map<tcaps_flow_feature_type,tcaps_flow_feature>::iterator ff_itr;
	
	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\nflow_state_changed: %d\n", flow_state_changed);
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::to_string() Acquiring flow_features_mutex\n"));
	#endif
	
	flow_features_mutex.acquire();
	
	#ifdef _MUTEX_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::to_string() flow_features_mutex acquired\n"));
	#endif
	
	for(ff_itr = flow_features.begin(); ff_itr != flow_features.end(); ff_itr++)
	{
		unsigned int padamt = strlen(flow_feature::get_flow_feature_description(ff_itr->first)) > 100 ? 1 : (100-strlen(flow_feature::get_flow_feature_description(ff_itr->first)));
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::to_string() 70-%d\tpadamt = %d\n",strlen(flow_feature::get_flow_feature_description(ff_itr->first)), padamt));
		snprintf(buf+strlen(buf), bufsize-strlen(buf), "\n%s:", flow_feature::get_flow_feature_description(ff_itr->first));
		pad(buf+strlen(buf), bufsize-strlen(buf), padamt);
		snprintf(buf+strlen(buf), bufsize-strlen(buf), "%0.2f", ff_itr->second);
	}
	
	flow_features_mutex.release();
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow::to_string() flow_features_mutex released\n"));
}

static void pad(char * string, unsigned int len, unsigned int n)
{
	for(unsigned int i = 0; i < n; i++)
	{
		snprintf(string+strlen(string), len-strlen(string), " ");
	}
} 



//void TCAPS_Flow::get_fc_ip(struct ip_address * ip)
//{
//	ip->ip_version = tcaps_fc_flow_belongs_to.ip_version;
//	
//	if(tcaps_fc_flow_belongs_to.ip_version == TCAPS_SP_IP_VERSION_4)
//	{
//		ip->octets = (u8 *)malloc(NUM_IPV4_OCTETS);
//		memcpy(ip->octets, tcaps_fc_flow_belongs_to.octets, NUM_IPV4_OCTETS);
//	}
//		
//	else if(tcaps_fc_flow_belongs_to.ip_version == TCAPS_SP_IP_VERSION_6)
//	{
//		ip->octets = (u8 *)malloc(NUM_IPV6_OCTETS);
//		memcpy(ip->octets, tcaps_fc_flow_belongs_to.octets, NUM_IPV6_OCTETS);
//	}
//}




























//void handle_first_packet(Pcap_Packet * packet)
//{
//	last_rx_packet = new Pcap_Packet(*packet);
//	
//	if(packet->get_layer2_frame()->get_layer3_packet()->get_layer3_protocol() == LAYER3_PROTOCOL_IPV4)
//	{
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet we are about to add is IPv4\n"));
//		
//		tcaps_ip_ntoa(clientip, ipbuf1, TRUE);
//		tcaps_ip_ntoa(((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip(), ipbuf2, TRUE);
//		
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) clientip: %s\tpacket src ip: %s\n", ipbuf1, ipbuf2));
//		
//		if(memcmp(clientip->octets, ((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV4_OCTETS) == 0)
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
//			
//			is_client2other_packet = true;
//			last_rx_client2other_packet = new Pcap_Packet(*packet);
//
//		}
//		else
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
//
//			last_rx_other2client_packet = new Pcap_Packet(*packet);
//
//		}
//	}
//	else //LAYER3_PROTOCOL_IPV6
//	{
//		if(memcmp(clientip->octets, ((IPv6_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV6_OCTETS) == 0)
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
//			
//			is_client2other_packet = true;
//			last_rx_client2other_packet = new Pcap_Packet(*packet);
//		}
//		else
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
//
//			last_rx_other2client_packet = new Pcap_Packet(*packet);
//		}
//	}
//	
//	
//
//}
//
//
//
//
//
//void TCAPS_Flow::add_packet_to_non_full_flow(Pcap_Packet * packet, const struct ip_address * clientip)
//{
//	char ipbuf1[60], ipbuf2[60];
//	bool is_client2other_packet = false;
//	Pcap_Packet *tmp_last_rx_client2other_packet, *tmp_last_rx_other2client_packet;
//	
//	if(num_packets_bidir_since_beginning == 0)
//	{
//		last_rx_packet = new Pcap_Packet(*packet);
//	}
//	else
//	{
//		delete last_rx_packet;
//		last_rx_packet = new Pcap_Packet(*flow_packets.back());
//	}
//
//	//we assume that the layer 3 protocol is IP and that this has been enforced elsewhere
//	
//	if(packet->get_layer2_frame()->get_layer3_packet()->get_layer3_protocol() == LAYER3_PROTOCOL_IPV4)
//	{
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet we are about to add is IPv4\n"));
//		
//		tcaps_ip_ntoa(clientip, ipbuf1, TRUE);
//		tcaps_ip_ntoa(((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip(), ipbuf2, TRUE);
//		
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) clientip: %s\tpacket src ip: %s\n", ipbuf1, ipbuf2));
//		
//		if(memcmp(clientip->octets, ((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV4_OCTETS) == 0)
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
//			
//			is_client2other_packet = true;
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_client2other_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//		}
//		else
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_other2client_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//		}
//	}
//	else //LAYER3_PROTOCOL_IPV6
//	{
//		if(memcmp(clientip->octets, ((IPv6_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV6_OCTETS) == 0)
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
//			
//			is_client2other_packet = true;
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_client2other_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//		}
//		else
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_other2client_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//		}
//	}
//	
//	//we're free to add the packet to the list
//	flow_packets.push_back(packet);
//	num_packets_in_window++;
//	
//				
//	//adjust "since flow beginning" parameters
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to enter calculate_since_beginning_flow_features()\n"));
//	calculate_since_beginning_flow_features(packet, is_client2other_packet);
//	
//	if(tmp_last_rx_client2other_packet != NULL)
//	{
//		delete last_rx_client2other_packet;
//		last_rx_client2other_packet = tmp_last_rx_client2other_packet;
//	}
//		
//	if(tmp_last_rx_other2client_packet != NULL)
//	{
//		delete last_rx_other2client_packet;
//		last_rx_other2client_packet = tmp_last_rx_other2client_packet;
//	}
//}
//
//
//void TCAPS_Flow::add_packet_to_full_flow(Pcap_Packet * packet, const struct ip_address * clientip)
//{
//	char ipbuf1[60], ipbuf2[60];
//	bool is_client2other_packet = false;
//	Pcap_Packet *tmp_last_rx_client2other_packet, *tmp_last_rx_other2client_packet;
//	
//	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) list.size() before delete: %d\n", flow_packets.size()));
//	
//	//Call delete the PCAP_packet * stored in the first list element
//	delete(*flow_packets.begin());
//	
//	//Erase the first list element itself
//	flow_packets.erase(flow_packets.begin());
//	
//	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) list.size() after delete: %d\n", flow_packets.size()));
//
//	
//	if(num_packets_bidir_since_beginning == 0)
//	{
//		last_rx_packet = new Pcap_Packet(*packet);
//	}
//	else
//	{
//		delete last_rx_packet;
//		last_rx_packet = new Pcap_Packet(*flow_packets.back());
//	}
//
//	//we assume that the layer 3 protocol is IP and that this has been enforced elsewhere
//	
//	if(packet->get_layer2_frame()->get_layer3_packet()->get_layer3_protocol() == LAYER3_PROTOCOL_IPV4)
//	{
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet we are about to add is IPv4\n"));
//		
//		tcaps_ip_ntoa(clientip, ipbuf1, TRUE);
//		tcaps_ip_ntoa(((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip(), ipbuf2, TRUE);
//		
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) clientip: %s\tpacket src ip: %s\n", ipbuf1, ipbuf2));
//		
//		if(memcmp(clientip->octets, ((IPv4_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV4_OCTETS) == 0)
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
//			
//			is_client2other_packet = true;
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_client2other_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//		}
//		else
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_other2client_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//		}
//	}
//	else //LAYER3_PROTOCOL_IPV6
//	{
//		if(memcmp(clientip->octets, ((IPv6_Packet*)(packet->get_layer2_frame()->get_layer3_packet()))->get_src_ip()->octets, NUM_IPV6_OCTETS) == 0)
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's SRC IP is the same as the client we are monitoring\n")); 
//			
//			is_client2other_packet = true;
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_client2other_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_client2other_packet = new Pcap_Packet(*packet);
//			}
//		}
//		else
//		{
//			ACE_DEBUG ((LM_DEBUG, "(%P|%t) The packet's DST IP is the same as the client we are monitoring\n")); 
//			
//			if(num_packets_bidir_since_beginning == 0)
//			{
//				last_rx_other2client_packet = new Pcap_Packet(*packet);
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//			else
//			{
//				//need to take a copy of the packet because there could be a nasty situation if we did     last_rx_client2other_packet = packet
//				//where we get lots of packets from one side of the connection and the object is deleted out of the list
//				//i.e. the last_rx_client2other_packet pointer gets stuffed
//				tmp_last_rx_other2client_packet = new Pcap_Packet(*packet);
//			}
//		}
//	}
//	
//	//we're free to add the packet to the list
//	flow_packets.push_back(packet);
//	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) list.size() after insert: %d\n", flow_packets.size()));
//	num_packets_in_window++;
//	
//				
//	//adjust "since flow beginning" parameters
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to enter calculate_since_beginning_flow_features()\n"));
//	calculate_since_beginning_flow_features(packet, is_client2other_packet);
//	
//	if(tmp_last_rx_client2other_packet != NULL)
//	{
//		delete last_rx_client2other_packet;
//		last_rx_client2other_packet = tmp_last_rx_client2other_packet;
//	}
//		
//	if(tmp_last_rx_other2client_packet != NULL)
//	{
//		delete last_rx_other2client_packet;
//		last_rx_other2client_packet = tmp_last_rx_other2client_packet;
//	}
//}









//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS, num_packets_bidir_since_beginning));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PACKETS, num_packets_client2other_since_beginning));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PACKETS, num_packets_other2client_since_beginning));
//	
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_BIDIR_MS, average_interarrival_time_bidir_ms));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS, average_interarrival_time_client2other_ms));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS, average_interarrival_time_other2client_ms));
//	
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_ACTIVE_TIME_SECS, ((unsigned int)ACE_OS::time()) - flow_start_time_secs));
//	
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PACKETS_PER_SECOND, divideIfNotZero((float)num_packets_bidir_since_beginning, flow_start_time_secs)));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PACKETS_PER_SECOND, divideIfNotZero((float)num_packets_client2other_since_beginning, flow_start_time_secs)));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PACKETS_PER_SECOND, divideIfNotZero((float)num_packets_other2client_since_beginning, flow_start_time_secs)));
//	
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PAYLOAD_BYTES, num_payload_bytes_bidir_since_beginning));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PAYLOAD_BYTES, num_payload_bytes_client2other_since_beginning));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PAYLOAD_BYTES, num_payload_bytes_other2client_since_beginning));
//	
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_SECOND, divideIfNotZero((float)num_payload_bytes_bidir_since_beginning,flow_start_time_secs)));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_SECOND, divideIfNotZero((float)num_payload_bytes_client2other_since_beginning,flow_start_time_secs)));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_SECOND, divideIfNotZero((float)num_payload_bytes_other2client_since_beginning,flow_start_time_secs)));
//	
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_PACKET, divideIfNotZero((float)num_payload_bytes_bidir_since_beginning,num_packets_bidir_since_beginning)));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET, divideIfNotZero((float)num_payload_bytes_client2other_since_beginning,num_packets_client2other_since_beginning)));
//	flow_features.insert(feature_entry(TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET, divideIfNotZero((float)num_payload_bytes_other2client_since_beginning,num_packets_other2client_since_beginning)));
	
