#include "tcaps_manager.h"

TCAPS_Client_Manager_Task::TCAPS_Client_Manager_Task(TCAPS_Network_Task * ci_comms, TCAPS_Network_Task * fs_comms)
  : n_threads_ (1)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_Manager_Task ctor 0x%x\n", (void *) this));
	
	ci_comms_task = ci_comms;
	fs_comms_task = fs_comms;
}

/* Take care of cleanup & tell the user we're going away.  */
TCAPS_Client_Manager_Task::~TCAPS_Client_Manager_Task (void)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_Manager_Task dtor 0x%x\n", (void *) this));

  /* Get our shutdown notification out of the queue and release it.  */
  ACE_Message_Block *message;

  /* Like the getq() in svc() below, this will block until a message
    arrives.  By blocking, we know that the destruction will be paused
    until the last thread is done with the message block.  */
  this->getq (message);
  message->release();
}

/* Open the object to do work.  Next, we activate the Task into the
  number of requested threads.  */
int TCAPS_Client_Manager_Task::open (void *unused)
{
  ACE_UNUSED_ARG (unused);

  return this->activate (THR_NEW_LWP, n_threads_);
}

/* Tell the user we're closing and invoke the baseclass' close() to
  take care of things.  */
int TCAPS_Client_Manager_Task::close (u_long flags)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_Manager_Task close 0x%x\n", (void *) this));
  return inherited::close (flags);
}

/* Our svc() method waits for work on the queue and then processes
  that work.  */
int TCAPS_Client_Manager_Task::svc (void)
{
	int retval = 0;
	ACE_Message_Block * block;
	ACE_Time_Value wait_for_message(0, 100); //0.1ms
	struct timespec sleep_timer;
	
	sleep_timer.tv_sec = 0;
	sleep_timer.tv_nsec = 500000000; //500ms
	
	while(1)
	{
		
		//let's perform some housekeeping on the client list
		perform_housekeeping();
		handle_packets();
		
		//let's check if there are any messages in the task's signalling queue
		retval = this->getq(block, &wait_for_message);
		
		if(retval != -1)
		{
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) There is a message in the signalling queue!\n"));
			
			if(block->msg_type() == ACE_Message_Block::MB_HANGUP)
        	{
        		ACE_DEBUG ((LM_DEBUG, "(%P|%t) We've been told to shutdown the TCAPS_Client_Manager_Task\n"));
          		this->putq(block);
          		break; //break out of infinite while
        	}
        	else
        	{
        		//No idea what this block is, but we don't care as it shouldn't be here!
        		block->release();
        	}
		}
		
		ACE_OS::nanosleep(&sleep_timer); //sleep for a bit to be nice to other processes running on the system
	}
	
	return 0;
}

void TCAPS_Client_Manager_Task::perform_housekeeping()
{
	unsigned int client_count=0, rule_count=0;
	std::map<struct mac_address,TCAPS_Client *,maccmp>::iterator c_itr;
	std::map<u16,TCAPS_Rule *>::iterator r_itr;
	TCAPS_Rule * tmprule;
	TCAPS_Client * tmpclient;
	TCAPS_Rule_List * client_rules;
	char clientbuf[256];
	char rulebuf[256];
	
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_Manager_Task::perform_housekeeping()\n"));
	
	client_list.acquire_mutex();		
	    		
	std::map<struct mac_address,TCAPS_Client *,maccmp> c_list = client_list.get_client_list();
	
	for(c_itr = c_list.begin(); c_itr != c_list.end(); c_itr++, client_count++)
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing client %d\n", client_count+1));
		
		tmpclient = ((*c_itr).second);
		tmpclient->to_string(clientbuf, sizeof(clientbuf));
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", clientbuf));
		
		//check if client has expired (not polled us for a while)
		if(((unsigned int)ACE_OS::time() - tmpclient->get_last_poll_time()) >= CLIENT_POLL_TIMEOUT_THRESHOLD)
		{
			//let's delete this client from the list'
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) This client has not polled us. Deleting the client\n"));
			struct mac_address tmpmac = tmpclient->get_client_mac();
			client_list.remove_client(&tmpmac);
		}
		else
		{
			client_rules = tmpclient->get_client_rules();
			
			client_rules->acquire_mutex();
			
			std::map<u16,TCAPS_Rule *> r_list = client_rules->get_rule_list();
			
			for(r_itr = r_list.begin(), rule_count = 0; r_itr != r_list.end(); r_itr++, rule_count++)
			{
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing rule %d\n", rule_count+1));
				
				tmprule = ((*r_itr).second);
				
				tmprule->to_string(rulebuf, sizeof(rulebuf));
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", rulebuf));
				
				//check if rule is about to time out
				if(tmprule->get_expiry_time() >= ((unsigned int)ACE_OS::time() - RULE_UPDATE_BEFORE_EXPIRY_TIME))
				{
					//need to send a RMP updating the rule to the client
				}
			}
			
			client_rules->release_mutex();
			
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished processing client %d\n\n\n", client_count+1));
		}
		
	}
	
	client_list.release_mutex();
}

void TCAPS_Client_Manager_Task::handle_packets()
{
	ACE_INET_Addr remote;
	char buf[BUFFER_SIZE];
	
	struct tcaps_sp_packet * packet = receive_packet(&remote);
    
	while(packet != NULL)
	{
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) handle_packets(): We are about to handle a packet\n"));
		
		switch(packet->packet_type)
		{
			case TCAPS_RMP_PACKET_TYPE:
					tcaps_sp_rmp_tostring((struct tcaps_sp_rmp *)packet->packet, buf, sizeof(buf));
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received a tcaps_sp RMP packet that looks like this:\n%s\n", buf));
					tcaps_sp_rmp_destroy((struct tcaps_sp_rmp *)packet->packet);
				break;
			case TCAPS_CRMP_PACKET_TYPE:
					tcaps_sp_crmp_tostring((struct tcaps_sp_crmp *)packet->packet, buf, sizeof(buf));
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received a tcaps_sp CRMP packet that looks like this:\n%s\n", buf));
					handle_crmp_packet((struct tcaps_sp_crmp * )packet->packet, &remote);
					tcaps_sp_crmp_destroy((struct tcaps_sp_crmp *)packet->packet);
				break;
			case TCAPS_AP_PACKET_TYPE:
					tcaps_sp_ap_tostring((struct tcaps_sp_ap *)packet->packet, buf, sizeof(buf));
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received a tcaps_sp AP packet that looks like this:\n%s\n", buf));
					tcaps_sp_ap_destroy((struct tcaps_sp_ap *)packet->packet);
				break;
			case TCAPS_TP_PACKET_TYPE:
					tcaps_sp_tp_tostring((struct tcaps_sp_tp *)packet->packet, buf, sizeof(buf));
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received a tcaps_sp TP packet that looks like this:\n%s\n", buf));
					tcaps_sp_tp_destroy((struct tcaps_sp_tp *)packet->packet);
				break;
			case TCAPS_PP_PACKET_TYPE:
					tcaps_sp_pp_tostring((struct tcaps_sp_pp *)packet->packet, buf, sizeof(buf));
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received a tcaps_sp PP packet that looks like this:\n%s\n", buf));
					handle_pp_packet((struct tcaps_sp_pp * )packet->packet, &remote);
					tcaps_sp_pp_destroy((struct tcaps_sp_pp *)packet->packet);
				break;
			default:
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received an unexpected tcaps_sp packet type on the client port\n"));
				break;
		}

		free(packet);
		
		packet = receive_packet(&remote);
	}
}

struct tcaps_sp_packet * TCAPS_Client_Manager_Task::receive_packet(ACE_INET_Addr * remote)
{
	ACE_Message_Block *message;
	struct tcaps_sp_packet * packet = NULL;
	
	message = ci_comms_task->receive_packet();
	
	if(message != NULL)
	{
		unsigned int num_serialized_bytes;
		
		ACE_OS::memmove ((char *) remote, message->rd_ptr(), sizeof (ACE_INET_Addr));
		message->rd_ptr (sizeof (ACE_INET_Addr));  // Move the rd_ptr() beyond the data.
		
		ACE_OS::memmove ((char *) &num_serialized_bytes, message->rd_ptr(), sizeof (unsigned int));
		message->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
		
		u8 serialized_bytes[num_serialized_bytes];
		
		ACE_OS::memmove ((char *) &serialized_bytes, message->rd_ptr(), num_serialized_bytes);
		message->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
		
		message->release();
		
		packet = (struct tcaps_sp_packet *)malloc(sizeof(struct tcaps_sp_packet));
		
		tcaps_sp_parse_packet((u8 *)serialized_bytes, num_serialized_bytes, packet);
	}
	
	return packet;
}

void TCAPS_Client_Manager_Task::handle_pp_packet(struct tcaps_sp_pp * packet, ACE_INET_Addr * remote)
{
	char ipbuf[60];
	
	struct ip_address clientip;
	clientip.octets = NULL;
	u16 clientport;
		
	clientport = remote->get_port_number();
		
	if(remote->get_type() == PF_INET)
	{
		clientip.ip_version = TCAPS_SP_IP_VERSION_4;
		clientip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
	}
	else if(remote->get_type() == PF_INET6)
	{
		clientip.ip_version = TCAPS_SP_IP_VERSION_6;
		clientip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
	}
	
	tcaps_ip_aton(remote->get_host_addr((char *)&ipbuf, sizeof(ipbuf)), &clientip);
	
	TCAPS_Client * client = client_list.get_client(&packet->identifier);
	
	if(TCAPS_PP_POLL_TYPE_REGISTER == packet->poll_type)
	{
		//we need to check if this client exists in our client_list, and add it if it doesn't
		
		if(client == NULL)
		{
			//new client
			
			client = new TCAPS_Client(&packet->identifier, &clientip, clientport);
			send_cmp(TCAPS_SP_CMP_OP_TYPE_MONITOR, client);
			client_list.add_client(client);
			
			char buf[1024];
			client->to_string(buf, sizeof(buf));
			printf("Client with the folllowing details just registered:\n%s\n\n", buf);
			//ack registration
		}
		else
		{
			//existing client, check ip address is still same
			client->set_client_ip(&clientip);
			client->set_client_port(clientport);
			client->set_last_poll_time((unsigned int)ACE_OS::time());
			//if the address has changed, we should notify flow sifter
		}
	}
	
	else if(TCAPS_PP_POLL_TYPE_UNREGISTER == packet->poll_type)
	{
		//we need to check if this client exists in our client_list, and remove it if it does
		
		if(client != NULL)
		{
			//existing client, remove from client list
			
			char buf[1024];
			client->to_string(buf, sizeof(buf));
			printf("Client with the folllowing details just unregistered:\n%s\n\n", buf);
			
			client_list.remove_client(&packet->identifier);
			send_cmp(TCAPS_SP_CMP_OP_TYPE_REMOVE, client);
		}
	}
	
	else if(TCAPS_PP_POLL_TYPE_POLL == packet->poll_type)
	{
		//we need to check if this client exists in our client_list
		
		if(client == NULL)
		{
			//new client, ask them to register
		}
		else
		{
			//existing client, check ip address is still same
			client->set_client_ip(&clientip);
			client->set_client_port(clientport);
			client->set_last_poll_time((unsigned int)ACE_OS::time());
		}
	}
	
	if(clientip.octets != NULL)
		free(clientip.octets);
}


void TCAPS_Client_Manager_Task::handle_crmp_packet(struct tcaps_sp_crmp * packet, ACE_INET_Addr * remote)
{
	TCAPS_Client * client = client_list.get_client(&packet->client_ip);
	
	if(client != NULL)
	{
		//we found the client this rule applies to
		
		if(TCAPS_RMP_OP_TYPE_CREATE == packet->partial_rmp->op_type)
		{
			TCAPS_Rule * tmprule = client->get_client_rules()->get_rule(packet->partial_rmp);
			
			if(tmprule == NULL)
			{
				tmprule = new TCAPS_Rule(packet->partial_rmp);
			
				//let's fill in the missing RMP fields that the flow classifier doesn't fill in
				struct mac_address tmp = client->get_client_mac();
				tmprule->set_identifier(&tmp);
				tmprule->set_rule_num(client->get_client_rules()->get_next_rule_num());
				tmprule->set_timeout(RULE_TIMEOUT_DEFAULT);
				
				send_rmp(tmprule, client);
				
				client->get_client_rules()->add_rule(tmprule);
			}
		}
		else if(TCAPS_RMP_OP_TYPE_DELETE == packet->partial_rmp->op_type)
		{
			//look for the rule and delete it if it exists
			//signal the client to remove the rule as well
			
			TCAPS_Rule * tmprule = client->get_client_rules()->get_rule(packet->partial_rmp);
			
			if(tmprule != NULL)
			{
				tmprule->set_op_type(TCAPS_RMP_OP_TYPE_DELETE);
				send_rmp(tmprule, client);
				client->get_client_rules()->remove_rule(tmprule->get_rule_num());
			}
		}
		else if(TCAPS_RMP_OP_TYPE_UPDATE == packet->partial_rmp->op_type)
		{
			//Find the existing rule and update it to reflect the new "Priority On Match"
			
			TCAPS_Rule * tmprule = client->get_client_rules()->get_rule(packet->partial_rmp);
			
			if(tmprule != NULL)
			{
				//update rmp fields
			}
		}
	}
}

void TCAPS_Client_Manager_Task::send_rmp(TCAPS_Rule * therule, TCAPS_Client * theclient)
{
	ACE_INET_Addr remote;
	
    struct ip_address clientip = theclient->get_client_ip();
    char ipbuf[64];
    tcaps_ip_ntoa(&clientip, ipbuf, TRUE);
    
    remote = ACE_INET_Addr(theclient->get_client_port(), ipbuf);
		
  	Packet_Block *message;
  	
  	unsigned int serialized_bytes = therule->get_serialized_bytes();


  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + serialized_bytes);
  	
  	message->copy ((const char *) &remote, sizeof(remote));
  	message->copy ((const char *) &serialized_bytes, sizeof(unsigned int));
  	message->copy ((const char *) therule->get_serialized(), serialized_bytes);


	ci_comms_task->send_packet(message);
}

void TCAPS_Client_Manager_Task::send_cmp(u8 cmp_op_type, TCAPS_Client * theclient)
{
	struct tcaps_sp_cmp * cmp_packet;
	ACE_INET_Addr remote;
	
	struct ip_address clientip = theclient->get_client_ip();
	char ipbuf[64];
	
	tcaps_ip_ntoa(&clientip, ipbuf, TRUE);
	
	if(TCAPS_SP_CMP_OP_TYPE_MONITOR == cmp_op_type)
	{
		struct ip_address fs_assigned_to;
		fs_assigned_to.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		
		//hard coded at the moment... but eventually could assign an FS randomly from an available pool
		fs_assigned_to.ip_version = TCAPS_SP_IP_VERSION_4;
		tcaps_ip_aton(TCAPS_FS_IP, &fs_assigned_to);
			    
	    remote = ACE_INET_Addr(TCAPS_FS_REMOTE_PORT, TCAPS_FS_IP);
	    
	    //assign the FS IP to the client
	    theclient->set_fs_ip(&fs_assigned_to);
	    free(fs_assigned_to.octets);
	}
	else if(TCAPS_SP_CMP_OP_TYPE_REMOVE == cmp_op_type)
	{
		//pull the associated FS IP from the client and use that to construct remote
		char fsipbuf[64];
		struct ip_address fsip = theclient->get_fs_ip();
		tcaps_ip_ntoa(&fsip, fsipbuf, TRUE);
		remote = ACE_INET_Addr(TCAPS_FS_REMOTE_PORT, fsipbuf);
	}
	
	cmp_packet = tcaps_sp_cmp_create(cmp_op_type, clientip.ip_version, ipbuf);
		
  	Packet_Block *message;


  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + cmp_packet->serialized_bytes);
  	
  	message->copy ((const char *) &remote, sizeof(remote));
  	message->copy ((const char *) &(cmp_packet->serialized_bytes), sizeof(unsigned int));
  	message->copy ((const char *) cmp_packet->serialized, cmp_packet->serialized_bytes);


	fs_comms_task->send_packet(message);
	
	tcaps_sp_cmp_destroy(cmp_packet);

}

//void send_packet(struct tcaps_sp_packet * packet, struct ip_address * recipient_ip, unsigned short recipien1t_port)
//{
//	char iptmp[60];
//	tcaps_ip_ntoa(recipient_ip, iptmp, TRUE);	    
//    ACE_INET_Addr remote(recipient_port, iptmp);
//		
//  	Packet_Block *message;
//  	u8 * serialized;
//  	unsigned int serialized_bytes;
//  	
//  	switch(packet->packet_type)
//  	{
//  		case TCAPS_RMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_rmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_rmp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_PP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_pp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_pp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_TP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_tp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_tp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_AP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_ap *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_ap *)packet->packet)->serialized;
//  			break;
//  		default:
//  			serialized_bytes = 0;
//  			serialized = 0;
//  			break;
//  	}
//
//  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + serialized_bytes);
//  	
//  	message->copy ((const char *) &remote, sizeof(remote));
//  	message->copy ((const char *) &serialized_bytes, sizeof(unsigned int));
//  	message->copy ((const char *) serialized, serialized_bytes);
//
//
//	ci_comms_task->send_packet(message);
//}
