
#include <unistd.h>

#include "ace/OS.h"
#include "ace/Log_Msg.h"
#include "ace/SOCK_Dgram.h"
#include "ace/INET_Addr.h"

#include <tcaps_sp.h>

//binarygen defines
#define BINARYGEN_CONVTYPE_FLOWID 1
#define BINARYGEN_CONVTYPE_STRING 2

bool parse_option_args(int conversion_type, int argc, char * argv[]);
int binarygen_main(int argc, char * argv[]);
void binarygen_usage(char * progname);

//packetgen defines
#define OUTER_DIMENSIONS 50
#define MIDDLE_DIMENSIONS 20
#define INNER_DIMENSIONS 1024
#define BUFFER_SIZE 1600
#define NUM_REQUIRED_ARGS (6+1)
#define NUM_REQUIRED_OPTS 3 //-i, -p and -t are all required


#define NUM_CUSTOM_AP_ARGS 5
#define NUM_CUSTOM_TP_ARGS 6

#define NUM_CUSTOM_PP_SEGMENT_ARGS 4
#define NUM_CUSTOM_FFP_SEGMENT_ARGS 2
#define NUM_CUSTOM_RMP_SEGMENT_ARGS 7
#define NUM_CUSTOM_CRMP_SEGMENT_ARGS NUM_CUSTOM_RMP_SEGMENT_ARGS

#define MIN_CUSTOM_CMP_ARGS 3
#define MIN_CUSTOM_PP_ARGS 2
#define MIN_CUSTOM_RMP_ARGS 7
#define MIN_CUSTOM_CRMP_ARGS 9
#define MIN_CUSTOM_FFP_ARGS 7


int packetgen_main(int argc, char * argv[]);


int parse_type_arg(char * optarg);

void send_packet(u8 * serialized, unsigned int serialized_bytes, char * recipient_ip, u16 recipient_port);

void send_ap(char * recipient_ip, u16 recipient_port, int num_primary_args, char * argv[], bool check_parse);
void send_tp(char * recipient_ip, u16 recipient_port, int num_primary_args, char * argv[], bool check_parse);
void send_cmp(char * recipient_ip, u16 recipient_port, int num_primary_args, char * argv[], bool check_parse);

void send_pp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse);
void send_rmp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse);
void send_crmp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse);
void send_ffp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse);

void do_parse_check(char * packet_name, struct tcaps_sp_packet * orig_packet);

void packetgen_usage(char * progname);
void packetgen_usage_custom(char * progname, int custom_type);

bool handle_custom_fixedargs(int packet_type, int argc, int * numprimaryargs, int * locationofprimaryargs, int offset);
bool handle_custom_variableargs(int packet_type, int argc, char * argv[], int * numprimaryargs, int * num_segments, char * args);
bool parse_segments(int remaining_args, int args_for_a_segment, int * num_segs, char * argv[], char * args);



int main(int argc, char * argv[])
{
	
	if(strstr(argv[0], "tcaps_sp_packetgen") != NULL)
	{
		return packetgen_main(argc, argv);
	}
	else if(strstr(argv[0], "tcaps_sp_binarygen") != NULL)
	{
		return binarygen_main(argc, argv);
	}
	else
	{
		printf("\n\nUnexpected program name.\nExpecting to be run as either tcaps_sp_packetgen or tcaps_sp_binarygen.\nYou will need to recompile this program with the new program name if you want it to work.\n\n"); 
	}
	
	return -1;
}

int binarygen_main(int argc, char * argv[])
{
	int ch, conversion_type = 0;
	bool error = true;
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) argc: %d\n", argc));

	
	while((ch = getopt(argc, argv, "sf:a")) != -1)
    {
		switch(ch)
		{
			case 'f':
				if(strncmp(optarg, "ipip", strlen("ipip")) == 0
					|| strncmp(optarg, "ipipport", strlen("ipipport")) == 0
						|| strncmp(optarg, "ipipspi", strlen("ipipspi")) == 0)
				{
					conversion_type = BINARYGEN_CONVTYPE_FLOWID;
				}
				else
					error = true;
				break;
				
			case 'a':
				error = parse_option_args(conversion_type, argc, argv);
			case 's':
				conversion_type = BINARYGEN_CONVTYPE_STRING;
			case '?':
			default:
				//error = true;
				break;
		         
		}
	}
	
	if(error || argc < 3)
	{
		binarygen_usage(argv[0]);
		return -1;
	}
	
	return 0;
}

void binarygen_usage(char * progname)
{
	printf("\n%s [{-f <type>} or {-s}] -a <args>\n\n", progname);
	
	printf("-f: Get the binary representation of a TCAPS flow id.\n\tValid types: ipip, ipipport, ipipspi\n\n");
	printf("-s: Get the binary representation of an ASCII string.\n\n");
	printf("-a: Define args for the requested option. Args must be in correct order as defined in this usage message\n");
	printf("\t-f ipip -a srcip dstip\n\t-f ipipport -a srcip srcport dstip dstport\n\t-f ipipspi -a srcip srcspi dstip dstspi\n\n");
	printf("\t-s -a string\n\n");
	puts("");
}

bool parse_option_args(int conversion_type, int argc, char * argv[])
{
	bool error = false;
	char hexbytes[1024];
	
	if(conversion_type == BINARYGEN_CONVTYPE_FLOWID)
	{
		struct ip_address tmpip;
		int dstip_argv_index = 5, num_ip_bytes = 0;
		
		//printf("We are parsing a TCAPS flow id\n");
		
		if(argc < 6)
			error = true;
		
		else if(strncmp(argv[2], "ipipport", strlen("ipipport")) == 0
					|| strncmp(argv[2], "ipipspi", strlen("ipipspi")) == 0)
		{
			if(argc < 8)
				error = true;
		}
		
		if(!error)
		{
			if(strstr(argv[4], ".") != NULL)
			{
				printf("We are parsing IPv4 addresses!\n");
				tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
				tmpip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				num_ip_bytes = NUM_IPV4_OCTETS;
			}
			else
			{
				printf("We are parsing IPv6 addresses!\n");
				tmpip.ip_version = TCAPS_SP_IP_VERSION_6;
				tmpip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				num_ip_bytes = NUM_IPV6_OCTETS;
			}

			tcaps_ip_aton(argv[4], &tmpip);
			printHexBytes(tmpip.octets, num_ip_bytes, hexbytes, sizeof(hexbytes), FALSE);
			
			if(strncmp(argv[2], "ipipport", strlen("ipipport")) == 0)
			{
				u16 srcport = strtol(argv[5], NULL, 10);
				printHexBytes((u8*)&srcport, sizeof(u16), hexbytes+strlen(hexbytes), sizeof(hexbytes), FALSE);
				dstip_argv_index++;
			}
			
			else if(strncmp(argv[2], "ipipspi", strlen("ipipspi")) == 0)
			{
				u32 srcspi = strtol(argv[5], NULL, 10);
				printHexBytes((u8*)&srcspi, sizeof(u32), hexbytes+strlen(hexbytes), sizeof(hexbytes), FALSE);
				dstip_argv_index++;
			}
			
			tcaps_ip_aton(argv[dstip_argv_index], &tmpip);
			printHexBytes(tmpip.octets, num_ip_bytes, hexbytes+strlen(hexbytes), sizeof(hexbytes), FALSE);
			
			if(strncmp(argv[2], "ipipport", strlen("ipipport")) == 0)
			{
				u16 dstport = strtol(argv[7], NULL, 10);
				printHexBytes((u8*)&dstport, sizeof(u16), hexbytes+strlen(hexbytes), sizeof(hexbytes), FALSE);
			}
			
			else if(strncmp(argv[2], "ipipspi", strlen("ipipspi")) == 0)
			{
				u32 dstspi = strtol(argv[7], NULL, 10);
				printHexBytes((u8*)&dstspi, sizeof(u32), hexbytes+strlen(hexbytes), sizeof(hexbytes), FALSE);
			}
			
			free(tmpip.octets);
		}
	}
	
	else if(conversion_type == BINARYGEN_CONVTYPE_STRING)
	{
		if(argc != 4)
			error = true;
			
		if(!error)
		{
			int num_chars = strlen(argv[3])+1;
			
			sprintf(hexbytes, "String to be converted: %s\\0\nConverted hex string: ", argv[3]);
			
			printHexBytes((u8*)argv[3], num_chars, hexbytes+strlen(hexbytes), sizeof(hexbytes), FALSE);
		}
	}
	
	if(!error)
		printf("%s\n", hexbytes);
	
	return error;
}


int packetgen_main(int argc, char * argv[])
{
	int ch, packet_type, numprimaryargs = 0, num_segments = 0, found_required_args = 0;
	bool error = false, custom = false, check_parse = false;
	char dstip[16];
	u16 dstport;
	int location_of_primary_args = 0;
	char * args = (char *)calloc(OUTER_DIMENSIONS*MIDDLE_DIMENSIONS, INNER_DIMENSIONS);
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) argc: %d\n", argc));
	
	while((ch = getopt(argc, argv, "i:d:t:pc")) != -1 && !error && !custom)
    {
		switch(ch)
		{
			case 't':
		 		packet_type = parse_type_arg(optarg);
		 		if(packet_type == TCAPS_NULL_PACKET_TYPE)
		 			error = true;
		 		found_required_args++;
				break;
			case 'i':
				strncpy(dstip, optarg, sizeof(dstip));
				found_required_args++;
				break;
			case 'd':
				dstport = strtol(optarg, NULL, 10);
				if(dstport == 0)
					error = true;
				found_required_args++;
				break;
			case 'c':
				custom = true;
				break;
			case 'p':
				check_parse = true;
				break;
			case '?':
			default:
		         error = true;
		}
	}
	
	if(error || argc < NUM_REQUIRED_ARGS || found_required_args != NUM_REQUIRED_OPTS)
	{
		packetgen_usage(argv[0]);
		return -1;
	}
	
	
	//test
	//char tmpbuf[50];
	//u16 number = *((u16 *)tcaps_atohex("ffff"));
	//printHexBytes(tcaps_atohex("ffffF"), 3, tmpbuf, sizeof(tmpbuf), FALSE);
	//printf("tmpbuf: %s\n", tmpbuf);
	//printf("number: %u\n", number);
	
	if(custom)
	{
		bool failed = false;
		
		if(TCAPS_AP_PACKET_TYPE == packet_type
			|| TCAPS_TP_PACKET_TYPE == packet_type
				|| TCAPS_CMP_PACKET_TYPE == packet_type)
		{
			if(check_parse)
				failed = handle_custom_fixedargs(packet_type, argc, &numprimaryargs, &location_of_primary_args, 2);
			else
				failed = handle_custom_fixedargs(packet_type, argc, &numprimaryargs, &location_of_primary_args, 1);
		}
		else
		{
			failed = handle_custom_variableargs(packet_type, argc, argv, &numprimaryargs, &num_segments, args);
		}
		
		if(failed)
		{
			packetgen_usage_custom(argv[0], packet_type);
			return -1;
		}
	}
	
	switch(packet_type)
	{
		//fixed arg packet types
		case TCAPS_AP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a AP packet to %s:%d\n\n", dstip, dstport));
			send_ap(dstip, dstport, numprimaryargs, argv+location_of_primary_args, check_parse);
			break;
		case TCAPS_TP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a TP packet to %s:%d\n\n", dstip, dstport));
			send_tp(dstip, dstport, numprimaryargs, argv+location_of_primary_args, check_parse);
			break;
		case TCAPS_CMP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a CMP packet to %s:%d\n\n", dstip, dstport));
			send_cmp(dstip, dstport, numprimaryargs, argv+location_of_primary_args, check_parse);
			break;

		//variable arg packet types
		case TCAPS_PP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a PP packet to %s:%d\n\n", dstip, dstport));
			send_pp(dstip, dstport, numprimaryargs, num_segments, args, check_parse);
			break;
		case TCAPS_RMP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a RMP packet to %s:%d\n\n", dstip, dstport));
			send_rmp(dstip, dstport, numprimaryargs, num_segments, args, check_parse);
			break;
		case TCAPS_FFP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a FFP packet to %s:%d\n\n", dstip, dstport));
			send_ffp(dstip, dstport, numprimaryargs, num_segments, args, check_parse);
			break;
		case TCAPS_CRMP_PACKET_TYPE:
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a CRMP packet to %s:%d\n\n", dstip, dstport));
			send_crmp(dstip, dstport, numprimaryargs, num_segments, args, check_parse);
			break;
	}

    free(args);
    
    return 0;
}

void packetgen_usage(char * progname)
{
	printf("\n%s -i <dstip> -d <dstport> -t <type> [-p] [-c <args>] [-a <args>]\n\n", progname);
	
	printf("-i: Destination IPv4 address in A.B.C.D format e.g. 172.16.50.70\n\n");
	printf("-d: Destination TCP/UDP port number as an integer e.g. 65000\n\n");
	printf("-t: TCAPS signalling protocol packet type to send.\n\tValid options: rmp, ap, tp, pp, ffp, crmp, cmp\n\n");
	printf("-p: Perform a parsing check to see if the packet type's parsing routines are functioning correctly\n\n");
	printf("-c: Define a custom TCAPS signalling protocol packet of type <type>. Args must be in correct order\n\n");
	printf("-a: Define custom segments to append to the TCAPS signalling protocol packet of type <type>. Only applicable to rmp, pp, ffp and crmp packet types. Args must be in correct order\n");
	puts("");
}

void packetgen_usage_custom(char * progname, int packet_type)
{
	packetgen_usage(progname);

	//fixed arg packet types
	if(TCAPS_AP_PACKET_TYPE == packet_type)
	{
		printf("\n\nAP custom args\n\n");
		printf("<args>: mac acktype errorcode rulenum transid\n\n");
		printf("mac: MAC address, with or without colon separators, case insensitive e.g. aB:ff:0f:45:32:dE or aBff0f4532dE\n\n");
	}
	
	else if(TCAPS_TP_PACKET_TYPE == packet_type)
	{
		printf("\n\nTP custom args\n\n");
		printf("<args>: transtype transid numrules ipversion serverip serverport\n\n");
	}
	
	else if(TCAPS_CMP_PACKET_TYPE == packet_type)
	{
		printf("\n\nCMP custom args\n\n");
		printf("<args>: optype clientipversion clientip\tOR\n\n");
		printf("<args>: optype clientipversion clientip {tm|fc} {tm|fc}ipversion {tm|fc}ip\tOR\n\n");
		printf("<args>: optype clientipversion clientip fcipversion fcip tmipversion tmip\tOR\n\n");
	}
	
	
	
	//variable arg packet types
	else if(TCAPS_PP_PACKET_TYPE == packet_type)
	{
		printf("\n\nPP custom args\n\n");
		printf("-c <args>: mac polltype\n\n");
		printf("mac: MAC address, with or without colon separators, case insensitive e.g. aB:ff:0f:45:32:dE or aBff0f4532dE\n\n");
		printf("polltype: 1 (TCAPS_PP_POLL_TYPE_REGISTER) or 2 (TCAPS_PP_POLL_TYPE_UNREGISTER) or 3 (TCAPS_PP_POLL_TYPE_POLL)\n\n");
		printf("-a <args>: key keylength value valuelength\n\n");
	}
	
	else if(TCAPS_FFP_PACKET_TYPE == packet_type)
	{
		printf("\n\nFFP custom args\n\n");
		printf("-c <args>: mac tmipversion tmip flowidtype whichipisclient clientipversion flowid\n\n");
		printf("mac: MAC address, with or without colon separators, case insensitive e.g. aB:ff:0f:45:32:dE or aBff0f4532dE\n\n");
		printf("-a <args>: feature_type feature_value\n\n");
	}
	
	else if(TCAPS_RMP_PACKET_TYPE == packet_type)
	{
		printf("\n\nRMP custom args\n\n");
		printf("-c <args>: mac rulenum transferid numretries optype matchpri timeout\n\n");
		printf("mac: MAC address, with or without colon separators, case insensitive e.g. aB:ff:0f:45:32:dE or aBff0f4532dE\n\n");
		printf("-a <args>: relto offset numbits2match matchrule conjunction refbits refbitslength\n\n");
	}
	
	else if(TCAPS_CRMP_PACKET_TYPE == packet_type)
	{
		printf("\n\nCRMP custom args\n\n");
		printf("-c <args>: clientipversion clientip mac rulenum transferid numretries optype matchpri timeout\n\n");
		printf("mac: MAC address, with or without colon separators, case insensitive e.g. aB:ff:0f:45:32:dE or aBff0f4532dE\n\n");
		printf("-a <args>: relto offset numbits2match matchrule conjunction refbits refbitslength\n\n");
	}
	
	
	puts("");
}

bool handle_custom_fixedargs(int packet_type, int argc, int * numprimaryargs, int * locationofprimaryargs, int offset)
{
	bool failed = false;
	int remainingargs = argc-NUM_REQUIRED_ARGS-offset; //-1 to include the -c
	*locationofprimaryargs = NUM_REQUIRED_ARGS+offset;
		
	if(packet_type == TCAPS_AP_PACKET_TYPE)
	{
		if(remainingargs-NUM_CUSTOM_AP_ARGS != 0)
		{
			failed = true;
		}
		else
		{
			*numprimaryargs = NUM_CUSTOM_AP_ARGS;
		}
	}
	
	else if(packet_type == TCAPS_TP_PACKET_TYPE)
	{
		if(remainingargs-NUM_CUSTOM_TP_ARGS != 0)
		{
			failed = true;
		}
		else
		{
			*numprimaryargs = NUM_CUSTOM_TP_ARGS;
		}
	}
	
	else if(packet_type == TCAPS_CMP_PACKET_TYPE)
	{
		if(remainingargs-MIN_CUSTOM_CMP_ARGS != 0
			&& remainingargs-MIN_CUSTOM_CMP_ARGS != 3 
				&& remainingargs-MIN_CUSTOM_CMP_ARGS != 4) //caters for standard ctor, ctor with either fm or fc ip, ctor with fm and fc ip
		{
			failed = true;
		}
		else
		{
			if(remainingargs-MIN_CUSTOM_CMP_ARGS == 0)
				*numprimaryargs = MIN_CUSTOM_CMP_ARGS;
				
			else if(remainingargs-MIN_CUSTOM_CMP_ARGS == 3)
				*numprimaryargs = MIN_CUSTOM_CMP_ARGS+3;
				
			else
				*numprimaryargs = MIN_CUSTOM_CMP_ARGS+4;
		}
	}
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Assigning numprimaryargs: %d\n\n", *numprimaryargs));
	
	return failed;
}





bool handle_custom_variableargs(int packet_type, int argc, char * argv[], int * numprimaryargs, int * num_segments, char * args)
{
	bool failed = false;
	int remainingargs = argc-NUM_REQUIRED_ARGS-1; //-1 to include the -c
	int locationofprimaryargs = NUM_REQUIRED_ARGS+1;
	
	if(packet_type == TCAPS_PP_PACKET_TYPE)
	{
		if(remainingargs-MIN_CUSTOM_PP_ARGS < 0)
		{
			failed = true;
		}
		else
		{
			strncpy(args, argv[NUM_REQUIRED_ARGS+1], INNER_DIMENSIONS);
			strncpy(args+INNER_DIMENSIONS, argv[NUM_REQUIRED_ARGS+2], INNER_DIMENSIONS);
			printf("primary pp args[0][0]: %s\n", args);
			printf("primary pp args[0][1]: %s\n", args+INNER_DIMENSIONS);
			
			*numprimaryargs = MIN_CUSTOM_PP_ARGS;
				
			if(remainingargs-MIN_CUSTOM_PP_ARGS == 0)
			{
				*num_segments = 0;
			}
			else
			{
				failed = parse_segments(remainingargs-*numprimaryargs, NUM_CUSTOM_PP_SEGMENT_ARGS, num_segments, argv+locationofprimaryargs+*numprimaryargs, args);
			}
		}
	}
	
	else if(packet_type == TCAPS_FFP_PACKET_TYPE)
	{
		if(remainingargs-MIN_CUSTOM_FFP_ARGS < 0)
		{
			failed = true;
		}
		else
		{
			strncpy(args, argv[NUM_REQUIRED_ARGS+1], INNER_DIMENSIONS);
			strncpy(args+INNER_DIMENSIONS, argv[NUM_REQUIRED_ARGS+2], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*2), argv[NUM_REQUIRED_ARGS+3], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*3), argv[NUM_REQUIRED_ARGS+4], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*4), argv[NUM_REQUIRED_ARGS+5], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*5), argv[NUM_REQUIRED_ARGS+6], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*6), argv[NUM_REQUIRED_ARGS+7], INNER_DIMENSIONS);

			printf("primary ffp args[0][0]: %s\n", args);
			printf("primary ffp args[0][1]: %s\n", args+INNER_DIMENSIONS);
			printf("primary ffp args[0][2]: %s\n", args+(INNER_DIMENSIONS*2));
			printf("primary ffp args[0][3]: %s\n", args+(INNER_DIMENSIONS*3));
			printf("primary ffp args[0][4]: %s\n", args+(INNER_DIMENSIONS*4));
			printf("primary ffp args[0][5]: %s\n", args+(INNER_DIMENSIONS*5));
			printf("primary ffp args[0][6]: %s\n", args+(INNER_DIMENSIONS*6));
			
			*numprimaryargs = MIN_CUSTOM_FFP_ARGS;
				
			if(remainingargs-MIN_CUSTOM_FFP_ARGS == 0)
			{
				*num_segments = 0;
			}
			else
			{
				failed = parse_segments(remainingargs-*numprimaryargs, NUM_CUSTOM_FFP_SEGMENT_ARGS, num_segments, argv+locationofprimaryargs+*numprimaryargs, args);
			}
		}
	}
	
	else if(packet_type == TCAPS_RMP_PACKET_TYPE)
	{
		if(remainingargs-MIN_CUSTOM_RMP_ARGS < 0)
		{
			failed = true;
		}
		else
		{
			strncpy(args, argv[NUM_REQUIRED_ARGS+1], INNER_DIMENSIONS);
			strncpy(args+INNER_DIMENSIONS, argv[NUM_REQUIRED_ARGS+2], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*2), argv[NUM_REQUIRED_ARGS+3], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*3), argv[NUM_REQUIRED_ARGS+4], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*4), argv[NUM_REQUIRED_ARGS+5], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*5), argv[NUM_REQUIRED_ARGS+6], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*6), argv[NUM_REQUIRED_ARGS+7], INNER_DIMENSIONS);

			printf("primary rmp args[0][0]: %s\n", args);
			printf("primary rmp args[0][1]: %s\n", args+INNER_DIMENSIONS);
			printf("primary rmp args[0][2]: %s\n", args+(INNER_DIMENSIONS*2));
			printf("primary rmp args[0][3]: %s\n", args+(INNER_DIMENSIONS*3));
			printf("primary rmp args[0][4]: %s\n", args+(INNER_DIMENSIONS*4));
			printf("primary rmp args[0][5]: %s\n", args+(INNER_DIMENSIONS*5));
			printf("primary rmp args[0][6]: %s\n", args+(INNER_DIMENSIONS*6));
			
			*numprimaryargs = MIN_CUSTOM_RMP_ARGS;
				
			if(remainingargs-MIN_CUSTOM_RMP_ARGS == 0)
			{
				*num_segments = 0;
			}
			else
			{
				failed = parse_segments(remainingargs-*numprimaryargs, NUM_CUSTOM_RMP_SEGMENT_ARGS, num_segments, argv+locationofprimaryargs+*numprimaryargs, args);
			}
		}
	}
	
	else if(packet_type == TCAPS_CRMP_PACKET_TYPE)
	{
		if(remainingargs-MIN_CUSTOM_CRMP_ARGS < 0)
		{
			failed = true;
		}
		else
		{
			strncpy(args, argv[NUM_REQUIRED_ARGS+1], INNER_DIMENSIONS);
			strncpy(args+INNER_DIMENSIONS, argv[NUM_REQUIRED_ARGS+2], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*2), argv[NUM_REQUIRED_ARGS+3], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*3), argv[NUM_REQUIRED_ARGS+4], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*4), argv[NUM_REQUIRED_ARGS+5], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*5), argv[NUM_REQUIRED_ARGS+6], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*6), argv[NUM_REQUIRED_ARGS+7], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*7), argv[NUM_REQUIRED_ARGS+8], INNER_DIMENSIONS);
			strncpy(args+(INNER_DIMENSIONS*8), argv[NUM_REQUIRED_ARGS+9], INNER_DIMENSIONS);

			printf("primary crmp args[0][0]: %s\n", args);
			printf("primary crmp args[0][1]: %s\n", args+INNER_DIMENSIONS);
			printf("primary crmp args[0][2]: %s\n", args+(INNER_DIMENSIONS*2));
			printf("primary crmp args[0][3]: %s\n", args+(INNER_DIMENSIONS*3));
			printf("primary crmp args[0][4]: %s\n", args+(INNER_DIMENSIONS*4));
			printf("primary crmp args[0][5]: %s\n", args+(INNER_DIMENSIONS*5));
			printf("primary crmp args[0][6]: %s\n", args+(INNER_DIMENSIONS*6));
			printf("primary crmp args[0][7]: %s\n", args+(INNER_DIMENSIONS*7));
			printf("primary crmp args[0][8]: %s\n", args+(INNER_DIMENSIONS*8));
			
			*numprimaryargs = MIN_CUSTOM_CRMP_ARGS;
				
			if(remainingargs-MIN_CUSTOM_CRMP_ARGS == 0)
			{
				*num_segments = 0;
			}
			else
			{
				failed = parse_segments(remainingargs-*numprimaryargs, NUM_CUSTOM_CRMP_SEGMENT_ARGS, num_segments, argv+locationofprimaryargs+*numprimaryargs, args);
			}
		}
	}
	


	return failed;
}




bool parse_segments(int remaining_args, int args_for_a_segment, int * num_segs, char * argv[], char * args)
{
	*num_segs = 0;
	bool failed = false;
	int tmp = 0;
	unsigned int index = MIDDLE_DIMENSIONS*INNER_DIMENSIONS; //we start after the first 2D array of 20*1024
	
	//printf("parse_segments()    remaining_args: %d\n", remaining_args);
	
	for(int i = 0; i < remaining_args && !failed; i++)
	{
		//printf("parse_segments()    argv[%d]: %s\n", i, argv[i]);
		if(strncmp(argv[i], "-a", 2) == 0 )
		{
			tmp = i+1;
			
			if(tmp <= (remaining_args-args_for_a_segment))
			{
				//we have found an append option
				//printf("We have found an append option!\n");
				
				(*num_segs)++;
				
				for(int j = 0; j < args_for_a_segment; j++)
				{
					//check to make sure that the args that should be for the segment aren't missing
					if(strncmp(argv[tmp+j], "-a", 2) != 0)
					{
						//printf("About to copy arg[%d]: %s\n", tmp+j, argv[tmp+j]);
						//strncpy(args+((*num_segs)*MIDDLE_DIMENSIONS)+(j*INNER_DIMENSIONS), argv[tmp+j], INNER_DIMENSIONS);
						
						strncpy(args+index+(j*INNER_DIMENSIONS), argv[tmp+j], INNER_DIMENSIONS);
					}
					else
					{
						//we stumbled across a -a part of the way through reading our expected number of args, which meanswe're missing an arg
						printf("Failed if(strncmp(argv[tmp+j], \"-a\", 2) != 0)\n");
						failed = true;
					}
				}
				
				index = index + (MIDDLE_DIMENSIONS*INNER_DIMENSIONS); //move to next boundary
			}
			else
			{
				printf("Failed if(tmp <= (remaining_args-args_for_a_segment))\n");
				failed = true;
			}
		}
	}
	
//	//if(!failed)
//	{
//		//char buf[INNER_DIMENSIONS*5];
//		
//		printf("failed: %d\nnum_segs: %d\n", failed, *num_segs);
//		
//		for(int i = 1; i <= (*num_segs); i++)
//		{
//			index = i*MIDDLE_DIMENSIONS*INNER_DIMENSIONS;
//			for(int j = 0; j < args_for_a_segment; j++)
//			{
//				printf("parsed arg %d: %s\n", i, args+index+(j*INNER_DIMENSIONS));
//				
//				//printHexBytes((const u8 *)args+index+(j*INNER_DIMENSIONS), INNER_DIMENSIONS*1, buf, sizeof(buf), TRUE);
//				//printf("serialized: \n\n%s\n\n", buf);
//			}
//		}
//	}
	
	return failed;
}



int parse_type_arg(char * optarg)
{
	int ptype = 0;
	
	//printf("optarg: %s\n", optarg);
	
	if(strncmp(optarg, "rmp", 3) == 0)
		ptype = TCAPS_RMP_PACKET_TYPE;
	else if(strncmp(optarg, "ap", 2) == 0)
		ptype = TCAPS_AP_PACKET_TYPE;
	else if(strncmp(optarg, "tp", 2) == 0)
		ptype = TCAPS_TP_PACKET_TYPE;
	else if(strncmp(optarg, "pp", 2) == 0)
		ptype = TCAPS_PP_PACKET_TYPE;
	else if(strncmp(optarg, "ffp", 3) == 0)
		ptype = TCAPS_FFP_PACKET_TYPE;
	else if(strncmp(optarg, "crmp", 4) == 0)
		ptype = TCAPS_CRMP_PACKET_TYPE;
	else if(strncmp(optarg, "cmp", 3) == 0)
		ptype = TCAPS_CMP_PACKET_TYPE;
	else
		ptype = TCAPS_NULL_PACKET_TYPE;
		
	return ptype;
}

void send_ap(char * recipient_ip, u16 recipient_port, int num_primary_args, char * argv[], bool check_parse)
{
	char buf[BUFFER_SIZE];
	struct tcaps_sp_ap * packet;
	
	if(num_primary_args > 0)
	{
		//printf("argv[0]: %s\n", argv[0]);
		
		packet = tcaps_sp_ap_create(argv[0],
									strtol(argv[1], NULL, 10),
									strtol(argv[2], NULL, 10),
									strtol(argv[3], NULL, 10),
									strtol(argv[4], NULL, 10));
	}
	else
	{
		packet = tcaps_sp_ap_create("123456789AFF", TCAPS_AP_ACK_TYPE_PP, 64899, 3, 254);
	}
		
    tcaps_sp_ap_tostring(packet, buf, sizeof(buf));
    	
    printf("AP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_ap_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("AP is valid\n");
    else
    	printf("AP is invalid\n");
    
    if(check_parse)
    {
    	struct tcaps_sp_packet ap_packet;
    	ap_packet.packet_type = TCAPS_AP_PACKET_TYPE;
    	ap_packet.packet = packet;
    	do_parse_check("AP", &ap_packet);
    }
    	
    send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}





void send_pp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse)
{
	char buf[BUFFER_SIZE];
	struct tcaps_sp_pp * packet;
	
	if(num_primary_args > 0)
	{
		int index = 0;
		
		printf("argv[0][0]: %s\n", args);
		
		packet = tcaps_sp_pp_create(args+0, strtol(args+INNER_DIMENSIONS, NULL, 10));
		
		for(int i = 1; i <= num_segments; i++)
		{
			index = i*MIDDLE_DIMENSIONS*INNER_DIMENSIONS;
			
			printf("Adding an info segment with key: %s, value: %s\n", args+index+(0*INNER_DIMENSIONS), args+index+(1*INNER_DIMENSIONS));
			
			tcaps_sp_pp_add_info_segment(packet, tcaps_sp_pp_create_info_segment((u8 *)args+index+(0*INNER_DIMENSIONS),
																				 strlen(args+index+(0*INNER_DIMENSIONS))+1,
																				 (u8 *)args+index+(1*INNER_DIMENSIONS),
																				 strlen(args+index+(1*INNER_DIMENSIONS))+1));
		}
	}
	else
	{
		packet = tcaps_sp_pp_create("123456789AFF", TCAPS_PP_POLL_TYPE_REGISTER);
		tcaps_sp_pp_add_info_segment(packet, tcaps_sp_pp_create_info_segment((u8 *)"muahaha evil one", strlen("muahaha evil one")+1, (u8 *)"model ZjX978Jg", strlen("model ZjX978Jg")+1));
	}
	
    tcaps_sp_pp_tostring(packet, buf, sizeof(buf));
    printf("PP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_pp_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("PP is valid\n");
    else
    	printf("PP is invalid\n");
    
    if(check_parse)
    {
    	struct tcaps_sp_packet pp_packet;
    	pp_packet.packet_type = TCAPS_PP_PACKET_TYPE;
    	pp_packet.packet = packet;
    	do_parse_check("PP", &pp_packet);
    }
    	
    send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}




void send_tp(char * recipient_ip, u16 recipient_port, int num_primary_args, char * argv[], bool check_parse)
{
	char buf[BUFFER_SIZE];
	
	struct tcaps_sp_tp * packet;
	
	if(num_primary_args > 0)
	{
		printf("argv[0]: %s\n", argv[0]);
		
		packet = tcaps_sp_tp_create(strtol(argv[0], NULL, 10),
									strtol(argv[1], NULL, 10),
									strtol(argv[2], NULL, 10),
									strtol(argv[3], NULL, 10),
									argv[4],
									strtol(argv[5], NULL, 10));
	}
	else
	{
		packet = tcaps_sp_tp_create(TCAPS_TP_TRANSFER_TYPE_RULE, 0, 60000, TCAPS_SP_IP_VERSION_4, "136.186.229.20", 65000);
	}
	
    tcaps_sp_tp_tostring(packet, buf, sizeof(buf));
    printf("TP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_tp_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("TP is valid\n");
    else
    	printf("TP is invalid\n");
    
    if(check_parse)
    {
    	struct tcaps_sp_packet tp_packet;
    	tp_packet.packet_type = TCAPS_TP_PACKET_TYPE;
    	tp_packet.packet = packet;
    	do_parse_check("TP", &tp_packet);
    }
    	
	send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}









void send_cmp(char * recipient_ip, u16 recipient_port, int num_primary_args, char * argv[], bool check_parse)
{
	char buf[BUFFER_SIZE];
	
	struct tcaps_sp_cmp * packet;
	
	if(num_primary_args > 0)
	{
		printf("argv[0]: %s\n", argv[0]);
		
		if(num_primary_args == MIN_CUSTOM_CMP_ARGS)
		{
			packet = tcaps_sp_cmp_create(strtol(argv[0], NULL, 10),
										 strtol(argv[1], NULL, 10),
										 argv[2]);
		}
		else if(num_primary_args == MIN_CUSTOM_CMP_ARGS+3)
		{
			if(strncmp(argv[3], "tm", 2) == 0)
			{
				packet = tcaps_sp_cmp_create_withtm(strtol(argv[0], NULL, 10),
											 		strtol(argv[1], NULL, 10),
											 		argv[2],
											 		strtol(argv[4], NULL, 10),
											 		argv[5]);
			}
			else if(strncmp(argv[3], "fc", 2) == 0)
			{
				packet = tcaps_sp_cmp_create_withfc(strtol(argv[0], NULL, 10),
											 		strtol(argv[1], NULL, 10),
											 		argv[2],
											 		strtol(argv[4], NULL, 10),
											 		argv[5]);
			}
		}
		else if(num_primary_args == MIN_CUSTOM_CMP_ARGS+4)
		{
			packet = tcaps_sp_cmp_create_withfctm(strtol(argv[0], NULL, 10),
										 		  strtol(argv[1], NULL, 10),
										 		  argv[2],
										 		  strtol(argv[3], NULL, 10),
										 		  argv[4],
										 		  strtol(argv[5], NULL, 10),
										 		  argv[6]);
		}
	}
	else
	{
		packet = tcaps_sp_cmp_create(TCAPS_SP_CMP_OP_TYPE_MONITOR, TCAPS_SP_IP_VERSION_4, "172.16.7.8");
	}
	
    tcaps_sp_cmp_tostring(packet, buf, sizeof(buf));
    printf("CMP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_cmp_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("CMP is valid\n");
    else
    	printf("CMP is invalid\n");
    
    if(check_parse)
    {
    	struct tcaps_sp_packet cmp_packet;
    	cmp_packet.packet_type = TCAPS_CMP_PACKET_TYPE;
    	cmp_packet.packet = packet;
    	do_parse_check("CMP", &cmp_packet);
    }
    	
	send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}







void send_ffp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse)
{
	char buf[BUFFER_SIZE];
	struct tcaps_sp_ffp * packet;
	
	if(num_primary_args > 0)
	{
		int index = 0;
		
		printf("argv[0][0]: %s\n", args);
		
		//char * mac, u8 tmipversion, char * tmip, u8 flowidtype, u8 whichipisclient, u8 clientipversion, const u8 * flowid
		packet = tcaps_sp_ffp_create(args+0,
									 strtol(args+INNER_DIMENSIONS, NULL, 10),
									 args+(2*INNER_DIMENSIONS),
									 strtol(args+(3*INNER_DIMENSIONS), NULL, 10),
									 strtol(args+(4*INNER_DIMENSIONS), NULL, 10),
									 strtol(args+(5*INNER_DIMENSIONS), NULL, 10),
									 tcaps_atohex(args+(6*INNER_DIMENSIONS)));
									 
		//printf("num_segments: %d\n", num_segments);
		
		for(int i = 1; i <= num_segments; i++)
		{
			index = i*MIDDLE_DIMENSIONS*INNER_DIMENSIONS;
			
			printf("Adding a feature segment with key: %s, value: %s\n", args+index+(0*INNER_DIMENSIONS), args+index+(1*INNER_DIMENSIONS));
			
			tcaps_sp_ffp_add_feature_segment(packet,
											 tcaps_sp_ffp_create_feature_segment(strtol((args+index+(0*INNER_DIMENSIONS)), NULL, 10), strtof((args+index+(1*INNER_DIMENSIONS)), NULL)));
		}
	}
	else
	{
		char hextmp[256];
		struct ip_address tmpip;
		tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
		tmpip.octets = (u8 *)malloc(4);
	
		
		//ipip flow_id
		char flowid1[8];
		tcaps_ip_aton("172.16.7.99", &tmpip);
		memcpy(&flowid1, tmpip.octets, 4);
		tcaps_ip_aton("172.16.7.100", &tmpip);
		memcpy(flowid1+4, tmpip.octets, 4);
		//free(tmpip.octets);
		printHexBytes((const u8*)flowid1, sizeof(flowid1), hextmp, sizeof(hextmp), FALSE);
		printf("Contents of ipip flowid in hex: %s\n", hextmp);
		
		//ipipport flow_id
		char flowid2[12];
		u16 src_port = 7654, dst_port = 4567;
		tcaps_ip_aton("172.16.7.89", &tmpip);
		memcpy(&flowid2, tmpip.octets, 4);
		tcaps_ip_aton("172.16.7.101", &tmpip);
		memcpy(flowid2+4, tmpip.octets, 4);
		//free(tmpip.octets);
		memcpy(flowid2+8, &src_port, sizeof(u16));
		memcpy(flowid2+10, &dst_port, sizeof(u16));
		printHexBytes((const u8*)flowid2, sizeof(flowid2), hextmp, sizeof(hextmp), FALSE);
		printf("Contents of ipipport flowid in hex: %s\n", hextmp);
		
		//ipipspi flow_id
		char flowid3[16];
		u32 src_spi = 4130376967, dst_spi = 3410279364;
		tcaps_ip_aton("172.16.7.99", &tmpip);
		memcpy(&flowid3, tmpip.octets, 4);
		tcaps_ip_aton("172.16.7.100", &tmpip);
		memcpy(flowid3+4, tmpip.octets, 4);
		//free(tmpip.octets);
		memcpy(flowid3+8, &src_spi, sizeof(u32));
		memcpy(flowid3+12, &dst_spi, sizeof(u32));
		printHexBytes((const u8*)flowid3, sizeof(flowid3), hextmp, sizeof(hextmp), FALSE);
		printf("Contents of ipipspi flowid in hex: %s\n", hextmp);
		
		free(tmpip.octets);
		
		
		char serialized_features[12];
		*((u16 *)serialized_features) = 1;
		*((float *)(serialized_features+sizeof(u16))) = 75.60;
		*((u16 *)(serialized_features+6)) = 2;
		*((float *)(serialized_features+6+sizeof(u16))) = 60.01;
		
		printHexBytes((const u8*)serialized_features, sizeof(serialized_features), hextmp, sizeof(hextmp), FALSE);
		printf("Contents of serialized_features in hex: %s\n", hextmp);
			
		//packet = tcaps_sp_ffp_create_withfeatures("123456789AFF", TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_FFP_FLOW_ID_TYPE_IPIP, TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT, TCAPS_SP_IP_VERSION_4, (u8 *)flowid1, (u8 *)serialized_features, sizeof(serialized_features)/6);
		packet = tcaps_sp_ffp_create_withfeatures("123456789AFF", TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT, TCAPS_FFP_FLOW_DST_IP_IS_CLIENT, TCAPS_SP_IP_VERSION_4, (u8 *)flowid2, (u8 *)serialized_features, sizeof(serialized_features)/6);
		//packet = tcaps_sp_ffp_create_withfeatures("123456789AFF", TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI, TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT, TCAPS_SP_IP_VERSION_4, (u8 *)flowid3, (u8 *)serialized_features, sizeof(serialized_features)/6);
	
	}
	
    tcaps_sp_ffp_tostring(packet, buf, sizeof(buf));
    printf("FFP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_ffp_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("FFP is valid\n");
    else
    	printf("FFP is invalid\n");
    
    if(check_parse)
    {
    	struct tcaps_sp_packet ffp_packet;
    	ffp_packet.packet_type = TCAPS_FFP_PACKET_TYPE;
    	ffp_packet.packet = packet;
    	do_parse_check("FFP", &ffp_packet);
    }
    	
    send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}






void send_rmp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse)
{
	char buf[BUFFER_SIZE];
	struct tcaps_sp_rmp * packet;
	
	if(num_primary_args > 0)
	{
		int index = 0;
		
		printf("argv[0][0]: %s\n", args);
		
		//char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout
		packet = tcaps_sp_rmp_create(args+0,
									 strtol(args+INNER_DIMENSIONS, NULL, 10),
									 strtol(args+(2*INNER_DIMENSIONS), NULL, 10),
									 strtol(args+(3*INNER_DIMENSIONS), NULL, 10),
									 strtol(args+(4*INNER_DIMENSIONS), NULL, 10),
									 strtol(args+(5*INNER_DIMENSIONS), NULL, 10),
									 strtol(args+(6*INNER_DIMENSIONS), NULL, 10));
									 
		//printf("num_segments: %d\n", num_segments);
		
		for(int i = 1; i <= num_segments; i++)
		{
			index = i*MIDDLE_DIMENSIONS*INNER_DIMENSIONS;
			
			//u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength
			printf("Adding a rule segment with\nrelto: %s\noffset: %s\nnumbits2match: %s\nmatchrule: %s\nconjunction: %s\nrefbits: %s\nrefbitslength: %s\n",
					args+index+(0*INNER_DIMENSIONS),
					args+index+(1*INNER_DIMENSIONS),
					args+index+(2*INNER_DIMENSIONS),
					args+index+(3*INNER_DIMENSIONS),
					args+index+(4*INNER_DIMENSIONS),
					args+index+(5*INNER_DIMENSIONS),
					args+index+(6*INNER_DIMENSIONS));
			
			tcaps_sp_rmp_add_rule_segment(packet,
										  tcaps_sp_rmp_create_rule_segment(strtol((args+index+(0*INNER_DIMENSIONS)), NULL, 10),
										 								   strtol((args+index+(1*INNER_DIMENSIONS)), NULL, 10),
										 								   strtol((args+index+(2*INNER_DIMENSIONS)), NULL, 10),
										 								   strtol((args+index+(3*INNER_DIMENSIONS)), NULL, 10),
										 								   strtol((args+index+(4*INNER_DIMENSIONS)), NULL, 10),
										 								   tcaps_atohex(args+index+(5*INNER_DIMENSIONS)),
										 								   strtol((args+index+(6*INNER_DIMENSIONS)), NULL, 10)));
		}
	}
	else
	{
		packet = tcaps_sp_rmp_create("123456789AFF", 45678, 130, 145, TCAPS_RMP_OP_TYPE_CREATE, 243, 55600);
		tcaps_sp_rmp_add_rule_segment(packet, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, 43000, 8, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8*)"abc", strlen("abc")+1));
		tcaps_sp_rmp_add_rule_segment(packet, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, 65522, 216, TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN, 245, (u8*)"stinky michelle eats my cds", strlen("stinky michelle eats my cds")+1));
	}
	
    tcaps_sp_rmp_tostring(packet, buf, sizeof(buf));
    printf("RMP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_rmp_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("RMP is valid\n");
    else
    	printf("RMP is invalid\n");
    
    if(check_parse)
    {
    	struct tcaps_sp_packet rmp_packet;
    	rmp_packet.packet_type = TCAPS_RMP_PACKET_TYPE;
    	rmp_packet.packet = packet;
    	do_parse_check("RMP", &rmp_packet);
    }
    
    send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}


void send_crmp(char * recipient_ip, u16 recipient_port, int num_primary_args, int num_segments, char * args, bool check_parse)
{
	char buf[BUFFER_SIZE];
	struct tcaps_sp_crmp * packet;
	
	if(num_primary_args > 0)
	{
		int index = 0;
		
		printf("argv[0][0]: %s\n", args);
		
		//u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout
		packet = tcaps_sp_crmp_create(strtol(args+0, NULL, 10),
									  args+INNER_DIMENSIONS,
									  args+(2*INNER_DIMENSIONS),
									  strtol(args+(3*INNER_DIMENSIONS), NULL, 10),
									  strtol(args+(4*INNER_DIMENSIONS), NULL, 10),
									  strtol(args+(5*INNER_DIMENSIONS), NULL, 10),
									  strtol(args+(6*INNER_DIMENSIONS), NULL, 10),
									  strtol(args+(7*INNER_DIMENSIONS), NULL, 10),
									  strtol(args+(8*INNER_DIMENSIONS), NULL, 10));
									 
		//printf("num_segments: %d\n", num_segments);
		
		for(int i = 1; i <= num_segments; i++)
		{
			index = i*MIDDLE_DIMENSIONS*INNER_DIMENSIONS;
			
			//u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength
			printf("Adding a rule segment with\nrelto: %s\noffset: %s\nnumbits2match: %s\nmatchrule: %s\nconjunction: %s\nrefbits: %s\nrefbitslength: %s\n",
					args+index+(0*INNER_DIMENSIONS),
					args+index+(1*INNER_DIMENSIONS),
					args+index+(2*INNER_DIMENSIONS),
					args+index+(3*INNER_DIMENSIONS),
					args+index+(4*INNER_DIMENSIONS),
					args+index+(5*INNER_DIMENSIONS),
					args+index+(6*INNER_DIMENSIONS));
			
			tcaps_sp_crmp_add_rule_segment(packet,
										   tcaps_sp_crmp_create_rule_segment(strtol((args+index+(0*INNER_DIMENSIONS)), NULL, 10),
										 								     strtol((args+index+(1*INNER_DIMENSIONS)), NULL, 10),
										 								     strtol((args+index+(2*INNER_DIMENSIONS)), NULL, 10),
										 								     strtol((args+index+(3*INNER_DIMENSIONS)), NULL, 10),
										 								     strtol((args+index+(4*INNER_DIMENSIONS)), NULL, 10),
										 								     tcaps_atohex(args+index+(5*INNER_DIMENSIONS)),
										 								     strtol((args+index+(6*INNER_DIMENSIONS)), NULL, 10)));
		}
	}
	else
	{
		packet = tcaps_sp_crmp_create(TCAPS_SP_IP_VERSION_4, "127.0.0.1", "000000000000", 45678, 130, 145, TCAPS_RMP_OP_TYPE_CREATE, 243, 55600);
		//packet = tcaps_sp_crmp_create(TCAPS_SP_IP_VERSION_4, "127.0.0.1", "123456789AFF", 45678, 130, 145, TCAPS_RMP_OP_TYPE_CREATE, 243, 55600);
		tcaps_sp_crmp_add_rule_segment(packet, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, 43000, 8, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8*)"abc", strlen("abc")+1));
		tcaps_sp_crmp_add_rule_segment(packet, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, 65522, 216, TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN, 245, (u8*)"stinky michelle eats my cds", strlen("stinky michelle eats my cds")+1));
	}
	
    tcaps_sp_crmp_tostring(packet, buf, sizeof(buf));
    printf("CRMP we are about to send looks like this:\n%s\n",buf);
    
    if(tcaps_sp_crmp_validate_serialized_packet(packet->serialized, packet->serialized_bytes))
    	printf("CRMP is valid\n");
    else
    	printf("CRMP is invalid\n");
    	
    if(check_parse)
    {
    	struct tcaps_sp_packet crmp_packet;
    	crmp_packet.packet_type = TCAPS_CRMP_PACKET_TYPE;
    	crmp_packet.packet = packet;
    	do_parse_check("CRMP", &crmp_packet);
    }
    	
    send_packet(packet->serialized, packet->serialized_bytes, recipient_ip, recipient_port);
}

void do_parse_check(char * packet_name, struct tcaps_sp_packet * orig_packet)
{
	struct tcaps_sp_packet parsed_packet;
	u8 * orig_serialized;
	unsigned int orig_serialized_bytes;
	char origbuf[BUFFER_SIZE];
	
	//printf("orig_packet->packet_type: %d\n", orig_packet->packet_type);
	
	switch(orig_packet->packet_type)
	{
		case TCAPS_RMP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_rmp *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_rmp *)(orig_packet->packet))->serialized;
			tcaps_sp_rmp_tostring(((struct tcaps_sp_rmp *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		case TCAPS_AP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_ap *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_ap *)(orig_packet->packet))->serialized;
			tcaps_sp_ap_tostring(((struct tcaps_sp_ap *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		case TCAPS_TP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_tp *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_tp *)(orig_packet->packet))->serialized;
			tcaps_sp_tp_tostring(((struct tcaps_sp_tp *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		case TCAPS_PP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_pp *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_pp *)(orig_packet->packet))->serialized;
			tcaps_sp_pp_tostring(((struct tcaps_sp_pp *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		case TCAPS_FFP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_ffp *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_ffp *)(orig_packet->packet))->serialized;
			tcaps_sp_ffp_tostring(((struct tcaps_sp_ffp *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		case TCAPS_CRMP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_crmp *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_crmp *)(orig_packet->packet))->serialized;
			tcaps_sp_crmp_tostring(((struct tcaps_sp_crmp *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		case TCAPS_CMP_PACKET_TYPE:
			orig_serialized_bytes = ((struct tcaps_sp_cmp *)(orig_packet->packet))->serialized_bytes;
			orig_serialized = ((struct tcaps_sp_cmp *)(orig_packet->packet))->serialized;
			tcaps_sp_cmp_tostring(((struct tcaps_sp_cmp *)(orig_packet->packet)), origbuf, sizeof(origbuf));
			break;
		default:
			break;
	}
    	
	tcaps_sp_parse_packet(orig_serialized, orig_serialized_bytes, &parsed_packet);
	
	if(parsed_packet.packet_type == TCAPS_NULL_PACKET_TYPE || parsed_packet.packet_type != orig_packet->packet_type)
	{
		printf("Parsing of %s type failed\n", packet_name);
	}
	else
	{
		char parsebuf[BUFFER_SIZE];
		bool same = false;
		
		//printf("parsed_packet->packet_type: %d\n", parsed_packet.packet_type);
		
		switch(parsed_packet.packet_type)
		{
			case TCAPS_RMP_PACKET_TYPE:
				same = tcaps_sp_rmp_compare_packets(((struct tcaps_sp_rmp *)(parsed_packet.packet)), ((struct tcaps_sp_rmp *)(orig_packet->packet)));
				tcaps_sp_rmp_tostring(((struct tcaps_sp_rmp *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			case TCAPS_AP_PACKET_TYPE:
				same = tcaps_sp_ap_compare_packets(((struct tcaps_sp_ap *)(parsed_packet.packet)), ((struct tcaps_sp_ap *)(orig_packet->packet)));
				tcaps_sp_ap_tostring(((struct tcaps_sp_ap *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			case TCAPS_TP_PACKET_TYPE:
				same = tcaps_sp_tp_compare_packets(((struct tcaps_sp_tp *)(parsed_packet.packet)), ((struct tcaps_sp_tp *)(orig_packet->packet)));
				tcaps_sp_tp_tostring(((struct tcaps_sp_tp *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			case TCAPS_PP_PACKET_TYPE:
				same = tcaps_sp_pp_compare_packets(((struct tcaps_sp_pp *)(parsed_packet.packet)), ((struct tcaps_sp_pp *)(orig_packet->packet)));
				tcaps_sp_pp_tostring(((struct tcaps_sp_pp *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			case TCAPS_FFP_PACKET_TYPE:
				same = tcaps_sp_ffp_compare_packets(((struct tcaps_sp_ffp *)(parsed_packet.packet)), ((struct tcaps_sp_ffp *)(orig_packet->packet)));
				tcaps_sp_ffp_tostring(((struct tcaps_sp_ffp *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			case TCAPS_CRMP_PACKET_TYPE:
				same = tcaps_sp_crmp_compare_packets(((struct tcaps_sp_crmp *)(parsed_packet.packet)), ((struct tcaps_sp_crmp *)(orig_packet->packet)));
				tcaps_sp_crmp_tostring(((struct tcaps_sp_crmp *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			case TCAPS_CMP_PACKET_TYPE:
				same = tcaps_sp_cmp_compare_packets(((struct tcaps_sp_cmp *)(parsed_packet.packet)), ((struct tcaps_sp_cmp *)(orig_packet->packet)));
				tcaps_sp_cmp_tostring(((struct tcaps_sp_cmp *)(parsed_packet.packet)), parsebuf, sizeof(parsebuf));
				break;
			default:
				break;
		}
		
			
		if(same)
		{
			printf("Parsing of %s type succeeded\n", packet_name);
		}
		else
		{
			printf("Parsing of %s type failed\n\nOriginal packet looks like this:\n%s\nParsed packet looks like this:\n%s\n", packet_name, origbuf, parsebuf);
		}
		
		
		switch(orig_packet->packet_type)
		{
			case TCAPS_RMP_PACKET_TYPE:
				tcaps_sp_rmp_destroy(((struct tcaps_sp_rmp *)(parsed_packet.packet)));
				break;
			case TCAPS_AP_PACKET_TYPE:
				tcaps_sp_ap_destroy(((struct tcaps_sp_ap *)(parsed_packet.packet)));
				break;
			case TCAPS_TP_PACKET_TYPE:
				tcaps_sp_tp_destroy(((struct tcaps_sp_tp *)(parsed_packet.packet)));
				break;
			case TCAPS_PP_PACKET_TYPE:
				tcaps_sp_pp_destroy(((struct tcaps_sp_pp *)(parsed_packet.packet)));
				break;
			case TCAPS_FFP_PACKET_TYPE:
				tcaps_sp_ffp_destroy(((struct tcaps_sp_ffp *)(parsed_packet.packet)));
				break;
			case TCAPS_CRMP_PACKET_TYPE:
				tcaps_sp_crmp_destroy(((struct tcaps_sp_crmp *)(parsed_packet.packet)));
				break;
			case TCAPS_CMP_PACKET_TYPE:
				tcaps_sp_cmp_destroy(((struct tcaps_sp_cmp *)(parsed_packet.packet)));
				break;
			default:
				break;
		}
		
	}
}


void send_packet(u8 * serialized, unsigned int serialized_bytes, char * recipient_ip, u16 recipient_port)
{
	ACE_INET_Addr remote(recipient_port, recipient_ip);

	ACE_INET_Addr local ((u_short) 0);
	ACE_SOCK_Dgram client;


	if(client.open(local) == -1)
    {
      ACE_DEBUG ((LM_DEBUG, "(%P|%t) ERROR! Unable to open UDP socket\n"));
    }
    else
    {
	    int retval = client.send (serialized, serialized_bytes, remote);
	
	    client.get_local_addr(local); 
	             
		ACE_DEBUG ((LM_DEBUG, "\n(%P|%t) Num bytes (%d) sent from (127.0.0.1:%d) to (%s:%d)\n", retval, local.get_port_number(), remote.get_host_addr(), remote.get_port_number()));
    }
}







/*

IPv6 address of www.kame.net
2001:200:0:8002:203:47ff:fea5:3085

RMP TEST CASES

tcaps_sp_rmp_create("123456789AFF", 45678, 130, 145, TCAPS_RMP_OP_TYPE_CREATE, 243, 55600);
tcaps_sp_rmp_add_rule_segment(pack2send, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, 43000, 8, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, "abc", strlen("abc")+1));
tcaps_sp_rmp_add_rule_segment(pack2send, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, 65522, 225, TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN, 245, "stinky michelle eats my cds", strlen("stinky michelle eats my cds")+1));


tcaps_sp_rmp_create_ipport2ipport("123456789AFF", 45678, 130, 145, TCAPS_RMP_OP_TYPE_CREATE, 0, 20, "10.10.10.10", 1010, "20.20.20.20", 2020, TCAPS_SP_IP_VERSION_4);
tcaps_sp_rmp_create_ipport2ipport("123456789AFF", 45678, 130, 145, TCAPS_RMP_OP_TYPE_DELETE, 0, 20, "10.10.10.10", 1010, "20.20.20.20", 2020, TCAPS_TP_IP_VERSION_4);



AP TEST CASES

tcaps_sp_ap_create("123456789AFF", TCAPS_AP_ACK_TYPE_PP, 64899, 3, 254);




PP TEST CASES

tcaps_sp_pp_create("123456789AFF", TCAPS_PP_POLL_TYPE_REGISTER);
    
tcaps_sp_pp_add_info_segment(pack2send, tcaps_sp_pp_create_info_segment("testkey", strlen("testkey")+1, "testvalue", strlen("testvalue")+1));
tcaps_sp_pp_add_info_segment(pack2send, tcaps_sp_pp_create_info_segment("testkey", strlen("testkey"), "testvalue", strlen("testvalue")));

tcaps_sp_pp_add_info_segment(pack2send, tcaps_sp_pp_create_info_segment("testkey2", strlen("testkey2")+1, "testvalue2", strlen("testvalue2")+1));
tcaps_sp_pp_add_info_segment(pack2send, tcaps_sp_pp_create_info_segment("yet another test", strlen("yet another test")+1, "7888956", strlen("7888956")+1));
tcaps_sp_pp_add_info_segment(pack2send, tcaps_sp_pp_create_info_segment("muahaha evil one", strlen("muahaha evil one")+1, "model ZjX978Jg", strlen("model ZjX978Jg")+1));
	
	
	
	
	

CMP TEST CASES
tcaps_sp_cmp_create(TCAPS_SP_CMP_OP_TYPE_MONITOR, TCAPS_SP_IP_VERSION_4, "172.16.7.8");
tcaps_sp_cmp_create_withfc(TCAPS_SP_CMP_OP_TYPE_MONITOR, TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_SP_IP_VERSION_4, "172.16.7.100");
tcaps_sp_cmp_create_withtm(TCAPS_SP_CMP_OP_TYPE_MONITOR, TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_SP_IP_VERSION_4, "172.16.7.100");
tcaps_sp_cmp_create_withfctm(TCAPS_SP_CMP_OP_TYPE_MONITOR, TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_SP_IP_VERSION_4, "172.16.7.100", TCAPS_SP_IP_VERSION_4, "172.16.7.150");


CRMP TEST CASES
tcaps_sp_crmp_create(TCAPS_SP_IP_VERSION_4, "127.0.0.1", "000000000000", 45678, 130, 145, TCAPS_RMP_OP_TYPE_CREATE, 243, 55600);
	

	
FFP TEST CASES

struct ip_address tmpip;
tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
tmpip.octets = (u8 *)malloc(4);



//ipip flow_id
char flowid1[8];
tcaps_ip_aton("172.16.7.99", &tmpip);
memcpy(&flowid1, tmpip.octets, 4);
tcaps_ip_aton("172.16.7.100", &tmpip);
memcpy(flowid1+4, tmpip.octets, 4);
//free(tmpip.octets);
printHexBytes((const u8*)flowid1, sizeof(flowid1), hextmp, sizeof(hextmp), FALSE);
printf("Contents of ipip flowid in hex: %s\n", hextmp);

//ipipport flow_id
char flowid2[12];
u16 src_port = 7654, dst_port = 4567;
tcaps_ip_aton("172.16.7.89", &tmpip);
memcpy(&flowid2, tmpip.octets, 4);
tcaps_ip_aton("172.16.7.101", &tmpip);
memcpy(flowid2+4, tmpip.octets, 4);
//free(tmpip.octets);
memcpy(flowid2+8, &src_port, sizeof(u16));
memcpy(flowid2+10, &dst_port, sizeof(u16));
printHexBytes((const u8*)flowid2, sizeof(flowid2), hextmp, sizeof(hextmp), FALSE);
printf("Contents of ipipport flowid in hex: %s\n", hextmp);

//ipipspi flow_id
char flowid3[16];
u32 src_spi = 4130376967, dst_spi = 3410279364;
tcaps_ip_aton("172.16.7.99", &tmpip);
memcpy(&flowid3, tmpip.octets, 4);
tcaps_ip_aton("172.16.7.100", &tmpip);
memcpy(flowid3+4, tmpip.octets, 4);
//free(tmpip.octets);
memcpy(flowid3+8, &src_spi, sizeof(u32));
memcpy(flowid3+12, &dst_spi, sizeof(u32));
printHexBytes((const u8*)flowid3, sizeof(flowid3), hextmp, sizeof(hextmp), FALSE);
printf("Contents of ipipspi flowid in hex: %s\n", hextmp);



free(tmpip.octets);


char serialized_features[12];
*((u16 *)serialized_features) = 1;
*((float *)(serialized_features+sizeof(u16))) = 75.60;
*((u16 *)(serialized_features+6)) = 2;
*((float *)(serialized_features+6+sizeof(u16))) = 60.01;

printHexBytes((const u8*)serialized_features, sizeof(serialized_features), hextmp, sizeof(hextmp), FALSE);
printf("Contents of serialized_features in hex: %s\n", hextmp);
	
tcaps_sp_ffp_create_withfeatures("123456789AFF", TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_FFP_FLOW_ID_TYPE_IPIP, TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT, TCAPS_SP_IP_VERSION_4, (u8 *)flowid1, (u8 *)serialized_features, sizeof(serialized_features)/6);
tcaps_sp_ffp_create_withfeatures("123456789AFF", TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT, TCAPS_FFP_FLOW_DST_IP_IS_CLIENT, TCAPS_SP_IP_VERSION_4, (u8 *)flowid2, (u8 *)serialized_features, sizeof(serialized_features)/6);
tcaps_sp_ffp_create_withfeatures("123456789AFF", TCAPS_SP_IP_VERSION_4, "172.16.7.8", TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI, TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT, TCAPS_SP_IP_VERSION_4, (u8 *)flowid3, (u8 *)serialized_features, sizeof(serialized_features)/6);
	
	
	
*/
